# Copyright 2016, 2017 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt


"""
netq - Query data across all nodes in fabric

Usage:
   netq add server <ip-server>|<server-name> [port <text-port>] [vrf <text-vrf>]
   netq (agent|notifier) (start|stop|status|restart)
   netq check (vlan|mtu) [unverified] [around <text-time>] [json]
   netq check agents [json]
   netq check (clag|ospf|vxlan|lnv|sensors|license) [around <text-time>] [json]
   netq check bgp [vrf <vrf>] [around <text-time>] [json]
   netq help [<text-keywords>]
   netq help verbose
   netq help list
   netq resolve [vrf <vrf>|vlan <1-4096>] [around <text-time>]
   netq [<hostname>] show agents [changes] [json]
   netq [<hostname>] show bgp [<ip-session>|asn <number-asn>] [vrf <vrf>] [around <text-time>] [json]
   netq <hostname> show bgp [<remote-interface>|<ip-session>] [vrf <vrf>] [around <text-time>] [json]
   netq [<hostname>] show bgp [<ip-session>|asn <number-asn>] [vrf <vrf>] changes [between <text-time> and <text-endtime>] [json]
   netq <hostname> show bgp [<remote-interface>|<ip-session>] [vrf <vrf>] changes [between <text-time> and <text-endtime>] [json]
   netq [<hostname>] show changes [<mac>|<ipv4>|<ipv6>|<ipv4> vrf <vrf>|<ipv6> vrf <vrf>|vrf <vrf>] between <text-time> and <text-endtime> [json]
   netq [<hostname>] show clag [around <text-time>] [json]
   netq [<hostname>] show clag changes [between <text-time> and <text-endtime>] [json]
   netq [<hostname>] show docker container [<ipv4> | <ipv6>] [portmap] [around <text-time>] [json]
   netq [<hostname>] show docker container [<ipv4> | <ipv6>] [portmap] changes [between <text-time> and <text-endtime>] [json]
   netq [<hostname>] show docker container <ipv4> | <ipv6> [network <network-name> | <ipv4/prefixlen>] [around <text-time>] [json]
   netq [<hostname>] show docker container <ipv4> | <ipv6> [network <network-name> | <ipv4/prefixlen>] changes [between <text-time> and <text-endtime>] [json]
   netq [<hostname>] show docker container [<ipv4> | <ipv6>] <proto> [<port>] [around <text-time>] [json]
   netq [<hostname>] show docker container [<ipv4> | <ipv6>] <proto> [<port>] [network <network-name> | <ipv4/prefixlen>] [around <text-time>] [json]
   netq [<hostname>] show docker container [<ipv4> | <ipv6>] <proto> [<port>] [network <network-name> | <ipv4/prefixlen>] changes [between <text-time> and <text-endtime>] [json]
   netq [<hostname>] show docker container [name <container-name> | image <container-image> | network <network-name> | mac <container-mac> ] [around <text-time>] [json]
   netq [<hostname>] show docker container [name <container-name> | image <container-image> | network <network-name> | mac <container-mac> ] changes [between <text-time> and <text-endtime>] [json]
   netq [<hostname>] show docker container [<ipv4> | <ipv6> | name <container-name> | image <container-image> | network <network-name> | mac <container-mac> ] connectivity [around <text-time>] [json]
   netq <hostname> show docker container adjacent [interfaces <remote-physical-interface>] [around <text-time>] [json]
   netq [<hostname>] show docker container name <container-name> adjacent [interfaces <remote-physical-interface>] [around <text-time>] [json]
   netq [<hostname>] show docker network [name <network-name> | <ipv4/prefixlen>] [brief] [around <text-time>] [json]
   netq [<hostname>] show docker network [name <network-name> | <ipv4/prefixlen>] [brief] changes [between <text-time> and <text-endtime>] [json]
   netq [<hostname>] show docker network driver <network-driver> [brief] [around <text-time>] [json]
   netq [<hostname>] show docker network driver <network-driver> [brief] changes [between <text-time> and <text-endtime>] [json]
   netq [<hostname>] show docker summary [<docker-version>] [around <text-time>] [json]
   netq [<hostname>] show docker summary [<docker-version>] changes [between <text-time> and <text-endtime>] [json]
   netq [<hostname>] show interfaces [around <text-time>] [count] [json]
   netq <hostname> show interfaces <remote-interface> [around <text-time>] [count] [json]
   netq [<hostname>] show interfaces type (bond|bridge|eth|loopback|macvlan|swp|vlan|vrf|vxlan) [around <text-time>] [count] [json]
   netq [<hostname>] show interfaces changes [between <text-time> and <text-endtime>] [json]
   netq <hostname> show interfaces <remote-interface> changes [between <text-time> and <text-endtime>] [json]
   netq [<hostname>] show interfaces type (bond|bridge|eth|loopback|macvlan|swp|vlan|vrf|vxlan) changes [between <text-time> and <text-endtime>] [json]
   netq [<hostname>] show lldp [changes]  [json]
   netq <hostname> show lldp <remote-physical-interface> [changes]  [json]
   netq [<hostname>] show lldp around <text-time> [json]
   netq <hostname> show lldp <remote-physical-interface> around <text-time> [json]
   netq [<hostname>] show lnv [around <text-time>] [json]
   netq [<hostname>] show lnv changes [between <text-time> and <text-endtime>] [json]
   netq [<hostname>] show macs [<mac>] [vlan <0-4096>] [origin] [around <text-time>] [json]
   netq [<hostname>] show macs [<mac>] [vlan <0-4096>] [around <text-time>] count [json]
   netq [<hostname>] show macs [<mac>] [vlan <0-4096>] [origin] changes [between <text-time> and <text-endtime>] [json]
   netq <hostname> show macs nexthop <interface-nexthop> [<mac>] [vlan <0-4096>] [origin] [around <text-time>] [json]
   netq <hostname> show macs nexthop <interface-nexthop> [<mac>] [vlan <0-4096>] [origin] changes [between <text-time> and <text-endtime>] [json]
   netq [<hostname>] show inventory brief [<inventory-search>|json] [json]
   netq [<hostname>] show inventory os [<os-version>|<os-name>|json] [json]
   netq [<hostname>] show inventory board [<board-vendor>|<board-model>|json] [json]
   netq [<hostname>] show inventory cpu [<cpu-arch>|<cpu-model>|json] [json]
   netq [<hostname>] show inventory disk [<disk-name>|<disk-size>|<disk-transport>|<disk-vendor>|json] [json]
   netq [<hostname>] show inventory asic [<asic-vendor>|<asic-model>|<asic-model-id>|json] [json]
   netq [<hostname>] show inventory memory [<memory-type>|<memory-size>|<memory-vendor>|json] [json]
   netq [<hostname>] show inventory license [cumulus] [around <text-time>] [json]
   netq [<hostname>] show ip addresses [<ipv4>|<ipv4/prefixlen>] [vrf <vrf>] [around <text-time>] [count] [json]
   netq <hostname> show ip addresses <remote-interface> [<ipv4>|<ipv4/prefixlen>] [vrf <vrf>] [around <text-time>] [count] [json]
   netq [<hostname>] show ip addresses [<ipv4>|<ipv4/prefixlen>] [vrf <vrf>] changes [between <text-time> and <text-endtime>] [json]
   netq <hostname> show ip addresses <remote-interface> [<ipv4>|<ipv4/prefixlen>] [vrf <vrf>] changes [between <text-time> and <text-endtime>] [json]
   netq [<hostname>] show ipv6 addresses [<ipv6>|<ipv6/prefixlen>] [vrf <vrf>] [around <text-time>] [count] [json]
   netq <hostname> show ipv6 addresses <remote-interface> [<ipv6>|<ipv6/prefixlen>] [vrf <vrf>] [around <text-time>] [count] [json]
   netq [<hostname>] show ipv6 addresses [<ipv6>|<ipv6/prefixlen>] [vrf <vrf>] [around <text-time>] [count] [json]
   netq <hostname> show ipv6 addresses <remote-interface> [<ipv6>|<ipv6/prefixlen>] [vrf <vrf>] [around <text-time>] [count] [json]
   netq [<hostname>] show ipv6 addresses [<ipv6>|<ipv6/prefixlen>] [vrf <vrf>] changes [between <text-time> and <text-endtime>] [json]
   netq <hostname> show ipv6 addresses <remote-interface> [<ipv6>|<ipv6/prefixlen>] [vrf <vrf>] changes [between <text-time> and <text-endtime>] [json]
   netq [<hostname>] show ip routes [<ipv4>|<ipv4/prefixlen>] [vrf <vrf>] [origin] [around <text-time>] [count] [json]
   netq [<hostname>] show ip routes [<ipv4>|<ipv4/prefixlen>] [vrf <vrf>] [origin] changes [between <text-time> and <text-endtime>] [json]
   netq [<hostname>] show ipv6 routes [<ipv6>|<ipv6/prefixlen>] [vrf <vrf>] [origin] [around <text-time>] [count] [json]
   netq [<hostname>] show ipv6 routes [<ipv6>|<ipv6/prefixlen>] [vrf <vrf>] [origin] changes [between <text-time> and <text-endtime>] [json]
   netq [<hostname>] show ip neighbors [<ipv4>|<ipv4> vrf <vrf>|vrf <vrf>] [<mac>]  [around <text-time>] [count] [json]
   netq <hostname> show ip neighbors <remote-interface> [<ipv4>|<ipv4> vrf <vrf>|vrf <vrf>] [<mac>] [around <text-time>] [count] [json]
   netq [<hostname>] show ip neighbors [<ipv4>|<ipv4> vrf <vrf>|vrf <vrf>] [<mac>] changes [between <text-time> and <text-endtime>] [json]
   netq <hostname> show ip neighbors <remote-interface> [<ipv4>|<ipv4> vrf <vrf>|vrf <vrf>] [<mac>] changes [between <text-time> and <text-endtime>] [json]
   netq [<hostname>] show ipv6 neighbors [<ipv6>|<ipv6> vrf <vrf>|vrf <vrf>] [<mac>] [around <text-time>] [count] [json]
   netq <hostname> show ipv6 neighbors <remote-interface> [<ipv6>|<ipv6> vrf <vrf>|vrf <vrf>] [<mac>] [around <text-time>] [count] [json]
   netq [<hostname>] show ipv6 neighbors [<ipv6>|<ipv6> vrf <vrf>|vrf <vrf>] [<mac>] changes [between <text-time> and <text-endtime>] [json]
   netq <hostname> show ipv6 neighbors <remote-interface> [<ipv6>|<ipv6> vrf <vrf>|vrf <vrf>] [<mac>] changes [between <text-time> and <text-endtime>] [json]
   netq [<hostname>] show sensors all [changes|around <text-time>] [json]
   netq [<hostname>] show sensors psu [<psu-name>] [changes|around <text-time>] [json]
   netq [<hostname>] show sensors temp [<temp-name>] [changes|around <text-time>] [json]
   netq [<hostname>] show sensors fan [<fan-name>] [changes|around <text-time>] [json]
   netq [<hostname>] show services [<service-name>] [vrf <vrf>] [active|monitored] [around <text-time>] [json]
   netq [<hostname>] show services [<service-name>] [vrf <vrf>] [active|monitored] changes [between <text-time> and <text-endtime>] [json]
   netq [<hostname>] show services [<service-name>] [vrf <vrf>] status (ok|warning|error|fail) [around <text-time>] [json]
   netq [<hostname>] show services [<service-name>] [vrf <vrf>] status (ok|warning|error|fail) changes [between <text-time> and <text-endtime>] [json]
   netq [<hostname>] show services [<service-name>] [vrf <vrf>] status (ok|warning|error|fail) changes [count] [json]
   netq <hostname> show stp topology [around <text-time>] [json]
   netq trace <mac> [vlan <1-4096>] from <hostname> [vrf <vrf>] [around <text-time>] [json]
   netq trace <ip> from (<hostname>|<ip-src>) [vrf <vrf>] [around <text-time>] [json]

Options:
   active         : Service is active
   add            : Add IP address or hostname of netq server
   addresses      : IPv4/v6 addresses
   agent          : Start/Stop/Restart/Query status of netq agent
   agents         : netq agent
   all            : Information about all subcomponents
   around         : Go back in time to around ...
   asic           : Network Processor
   asn            : BGP ASN
   adjacent       : Adjacent to the node
   between        : Specify start and end times for output desired
   bgp            : BGP info
   board          : Switch Board
   bond           : Name of bond on device
   bridge         : Name of bridge on device
   brief          : Brief summary
   changes        : How this infomation has changed with time
   check          : Perform fabric-wide checks
   clag           : Multi-chassis LAG (CLAG) info
   cluster        : Swarm Cluster
   connectivity   : Connectivity
   config         : Configuration
   container      : Container information
   count          : Count of matching entries
   cpu            : Management Processor
   cumulus        : Cumulus Linux
   disk           : Storage Disk
   docker         : Docker Info
   error          : Service has errors
   eth            : ethX interface
   example        : Show examples of usage and workflow
   fail           : Service has failures
   fan            : Fan sensor
   from           : Starting at node...
   help           : Show usage info
   id             : Id
   interfaces     : Network interface
   ip             : IPv4 related info
   ipv6           : IPv6 related info
   json           : Provide output in JSON
   image          : Image name
   impact         : Impact due to node outage
   info           : Show information available about ...
   inventory      : Inventory information
   keys           : View command keys
   l2             : Bridged path
   l3             : Routed path
   license        : License
   list           : Show all help commands
   lldp           : LLDP based neighbor info
   lnv            : Lightweight Network Virtualization info
   loopback       : Loopback interface
   mac            : Mac address
   macs           : Mac table entries
   macvlan        : Show mac vlan info
   memory         : Hardware RAM
   monitored      : Service output is analyzed by netq
   mode           : Docker service mode
   mtu            : Link MTU
   name           : Name
   neighbor       : IP neighbor cache (ARP/ND) info across the fabric
   neighbors      : IP neighbor info (ARP/ND)
   network        : Network Information
   nexthop        : Next Hop
   driver         : Docker network driver
   node           : Docker swarm node
   notifier       : Start/Stop/Restart/Query status of netq notifier
   ntp            : Network Time Protocol
   origin         : Owner of route
   ok             : Service is ok
   os             : Operating System
   ospf           : OSPF info
   port           : TCP port of DB server
   portmap        : Check Docker Container exposed ports
   protocol       : IP Protocol
   psu            : Power supply sensor
   quagga         : Quagga
   resolve        : Annotate input with names and interesting info
   restart        : Restart daemon
   routes         : IPv4/v6 routes
   rx             : Receive
   services       : System services
   service        : Docker Service
   show           : Show fabric-wide info about specified object
   start          : Start daemon
   startup        : Netq startup
   stop           : Stop daemon
   stp            : Spanning Tree
   status         : Display daemon status
   sensors        : Temperature/Fan/PSU sensors
   server         : Specify IP address of DB server
   summary        : Summary
   swarm          : Docker Swarm
   swp            : swpX interface
   temp           : Temperature sensor
   topology       : Active STP topology
   trace          : Control plane trace path across fabric
   tx             : Transmit
   type           : Interface type
   uptime         : Uptime
   unverified     : Show also the unverifiable interfaces
   verbose        : More help on help
   view           : Show output of pre-defined commands on specific node
   vlan           : VLAN
   vrf            : VRF
   vxlan          : VxLAN dataplane info
   warning        : Service has warnings
   server         : IP address of DB server
   <0-4096>       : VLAN ID
   <1-4096>       : VLAN ID, between 1-4096
   <asic-vendor>  : Network ASIC Vendor
   <asic-model>   : Network ASIC Model
   <asic-model-id> : Network ASIC Model ID
   <board-vendor> : Board Vendor
   <board-model>  : Board Model
   <cpu-arch>     : CPU Architecture
   <cpu-model>    : CPU Processor Model
   <cluster-id>   : Cluster Id
   <cluster-name> : Cluster name
   <container-id> : Container Id
   <container-image> : Container image name
   <container-mac> : Container mac
   <container-name> : Container Name
   <disk-name>    : Disk Name
   <disk-size>    : Disk Size
   <disk-transport> : Disk Transport
   <disk-vendor>  : Disk Vendor
   <docker-version> : Docker engine version
   <example>      : name of example
   <fan-name>     : Fan sensor name
   <hostname>     : Name of the remote node you want to query
   <interface-nextHop> : The name of the nexthop interface
   <inventory-search> : Search inventory for e.g., dell, trident2, x86, qsfp28, ddr3
   <ip-server>    : IP address of DB server
   <ip-session>   : IP address of BGP session (numbered)
   <ipv4>         : IPv4 address (no mask)
   <ipv6>         : IPv6 address (no mask)
   <ip>           : IPv4 or v6 address (no mask)
   <ip-src>       : Source IP address used in trace
   <ipv4/prefixlen> : IPv4 address with mask
   <ipv6/prefixlen> : IPv6 address with mask
   <ip/prefixlen> : IPv4 or IPv6 address with mask
   <mac>          : MAC address
   <memory-type>  : Memory Type
   <memory-size>  : Memory Size
   <memory-vendor> : Memory Vendor
   <mode>         : Mode
   <network-name> : Network name
   <network-driver> : Network Driver
   <node-id>      : Node Id
   <number-asn>     : BGP ASN
   <remote-interface> : Name of the interface on the remote node
   <remote-physical-interface> : Name of the physical interface on the remote node
   <service-name> : Name of service
   <service-id>   : Service id
   <swarm-service-name> : Name of Docker Swarm service
   <os-version>   : Operating System version
   <os-name>      : Operating System name
   <port>         : L4 Port
   <proto>        : IP Protocol
   <psu-name>     : Power supply sensor name
   <role>         : Docker Swarm Node Role (worker, master)
   <text-brname>  : Name of the bridge
   <text-time>    : Time string such as 10m, 30s, 1h
   <text-endtime> : Time string such as 10m, 30s, 1h
   <text-keywords>: help about one or more keywords in the commands
   <server-name>  : Host name of DB server (must be DNS resolvable)
   <temp-name>    : Temperature sensor name
   <text-port>    : Backend Port
   <text-vrf>     : Backend VRF
   <text-vlan>    : VLAN
   <vrf>          : VRF name

Help:
    -h --help     Show this screen

"""
from __future__ import absolute_import

import bisect
import functools
import itertools
import json
import os
import socket
import subprocess
import sys
import traceback
from datetime import datetime
from operator import itemgetter
import syslog
import logging
import re
from glob import glob

import tabulate
from network_docopt import (
    NetworkDocopt,
    get_network_docopt_info
)

from netq_lib.common import utils
from netq_lib.common.enums import (
    BgColors,
    DiffState,
    Heartbeat,
    ConfKw,
    ServiceStatus,
    ServiceStatusString,
    Platform
)
from netq_apps.lib.netq import (
    NetQ,
    NetQException,
    AgentHealth
)
from netq_apps.lib import completers

from netq_lib.orm.redisdb import (
    set_connection_settings,
    close_connection,
    ping_server,
    BackendExceptions,
)

_JSON = None
_EMPTY_JSON = '[{}]'
_RC_SUCCESS = 0
_RC_FAIL = 1
_PROC_NAME = 'netq-cii'
args = None

class _MyTabulate(object):
    @staticmethod
    def __line_row(colsizes, *_):
        return ' '.join('-' * colsize for colsize in colsizes)

    @classmethod
    def __build_row(cls, colsizes, sep, cell_values, _, colaligns):
        values_with_attrs = []
        new_cell_values = []
        for cell_value, colsize, colalign in zip(cell_values,
                                                 colsizes,
                                                 colaligns):
            color = None
            if colalign == 'right':
                fmt = '{0:>%d}' % colsize
            else:
                fmt = '{0:<%d}' % colsize
            key = cell_value.strip()
            if key.startswith(BgColors.RED):
                color = BgColors.RED
                key = fmt.format(utils.strip_colors(key).strip())
            elif key.startswith(BgColors.YELLOW):
                color = BgColors.YELLOW
                key = fmt.format(utils.strip_colors(key).strip())
            elif key.startswith(BgColors.GREEN):
                color = BgColors.GREEN
                key = fmt.format(utils.strip_colors(key).strip())
            else:
                key = fmt.format(key)
            if len(key) > colsize:
                idx = key.rfind(sep, 0, colsize)
                if idx == -1:
                    idx = colsize - 1
                if color:
                    new_cell_values.append(utils.color_wrap(color,
                                                            key[idx + 1:]))
                else:
                    new_cell_values.append(key[idx + 1:])
                key = key[:idx + 1] + ' ' * (colsize - idx - 1)
            else:
                new_cell_values.append('')
            if color:
                values_with_attrs.append(utils.color_wrap(color, key))
            else:
                values_with_attrs.append(key)
        line = ' '.join(values_with_attrs).rstrip()
        if any(new_cell_values):
            return line + '\n' + cls.__build_row(colsizes, sep,
                                                 new_cell_values, _,
                                                 colaligns)
        else:
            return line

    @classmethod
    def paginate(cls, data, headers, colsizes, sep=', ', first=False):
        colsizes = [max(x, len(y)) for x, y in zip(colsizes, headers)]
        linebelowheader = functools.partial(cls.__line_row, colsizes)
        headerrow = datarow = functools.partial(cls.__build_row, colsizes, sep)
        simple = tabulate.TableFormat(lineabove=None,
                                      linebelowheader=linebelowheader,
                                      linebetweenrows=None,
                                      linebelow=None,
                                      headerrow=headerrow,
                                      datarow=datarow,
                                      padding=0,
                                      with_header_hide=['lineabove',
                                                        'linebelow'])
        if first:
            return tabulate.tabulate(data, headers=headers, tablefmt=simple)
        else:
            return tabulate.tabulate(data, headers=[], tablefmt=simple)


def _build_diff(entries, state, mods):
    output = []
    for entry in entries:
        if entry is None:
            state += 1
            continue
        state_str = DiffState.MAP[state]
        if state == DiffState.MODIFIED:
            mods.append(entry)
            if len(mods) == 2:
                if mods[0][:-1] != mods[1][:-1]:
                    output.append((state_str,) + tuple(mods[0]))
                    output.append(('',) + tuple(mods[1]))
                mods[:] = []
        else:
            output.append((state_str,) + tuple(entry))
    return output, state


def _tabulate(output, headers=None, jsonify=False, first=True, colsizes=None):
    if jsonify:
        global _JSON  # pylint: disable=global-statement
        out = []
        sanitized_headers = []
        _JSON = _JSON or []
        for header in headers:
            if header.startswith('#'):
                header = re.sub(r'#', 'num', header)
            sanitized_headers.append(
                ''.join(i.lower() if idx == 0 else i.lower().title()
                        for idx, i in enumerate(header.split(' ')))
            )
        for ele in output:
            if 'Intf' in ele:
                d = {}
                for key, item in zip(sanitized_headers, ele):
                    if isinstance(item, basestring):
                        tmp = re.sub(r'\bIntf\b', '0', utils.strip_colors(item))
                        d.update({key: int(tmp) if tmp == "0" else tmp})
                    else:
                        d.update({key: item})
                out.append(d)
            else:
                out.append(dict((key, utils.strip_colors(item))
                                for key, item in zip(sanitized_headers, ele)))
        _JSON.append(out)
    else:
        # Stringify timestamps
        new_output = []
        for row in output:
            new_ele = []
            for ele in row:
                if isinstance(ele, float):
                    try:
                        datetime.fromtimestamp(ele)
                        new_ele.append(utils.pretty_date(ele))
                        continue
                    except Exception:  # pylint: disable=broad-except
                        pass
                new_ele.append(ele)
            new_output.append(new_ele)
        if colsizes is not None:
            p_out = _MyTabulate.paginate(new_output, headers=headers,
                                         colsizes=colsizes, first=first)
        else:
            p_out = tabulate.tabulate(new_output, headers)
        if first and p_out:
            print(p_out)
        elif p_out:
            print(p_out[p_out.find('\n', p_out.find('\n') + 1) + 1:])


def get_examples():
    '''Get list of all example files'''

    choices = []
    options = ''
    for example_dir in ('/usr/local/share/netq/example/',
                        '/usr/share/netq/example/'):

        if os.path.exists(example_dir):
            # traverse root directory, and list directories as dirs and
            # files as files
            for root, dirs, files in os.walk(example_dir):
                root = re.sub(example_dir, '', root)
                # We have the first line of the file be the help
                # string for the example
                for filename in sorted(files):
                    key = (root + '/' + filename).replace('/', ' ')
                    choices.append(key)
                    try:
                        options += '\n%s: %s' % (
                            key,
                            open(example_dir + '/' + root + '/' + filename, 'r').readline().strip())
                    except Exception:
                        pass
    return sorted(choices), options


def _add_dynamic_cmds(doc):
    """ Hack to remove <wildcard> hacks from network docopt
    """
    usage_idx = doc.find('Usage:')
    options_idx = doc.find('Options:')
    pre_usage = doc[:usage_idx]
    usage = doc[usage_idx:options_idx].rstrip('\n')
    post_usage = doc[options_idx:]

    eg_choices, eg_options = get_examples()
    eg_string = ('   netq example {key}')

    # try:
    #     view_choices = NetQ.get_view_options('*')
    #     view_aroundstring = ' around <text-time>'
    #     view_string = ('   netq <hostname> view {key}')
    # except Exception as ex:
    #     view_choices = []
    #     view_aroundstring = ''
    #     view_string = ''
    view_choices = []
    view_aroundstring = ''
    view_string = ''

    new_usage = '%s\n%s\n%s\n' % (usage,
                                  '\n'.join([view_string.format(key=choice)
                                             for choice in view_choices]),
                                  '\n'.join([view_string.format(key=choice) +
                                             view_aroundstring
                                             for choice in view_choices]) +
                                  '\n'.join([eg_string.format(key=choice)
                                             for choice in eg_choices]))

    return pre_usage + new_usage + post_usage + eg_options


# These cli_show* routines never print the JSON output. This allows the
# JSON output to be aggregated. This is done already with 'show changes'
# command


def cli_show_links(cli, netq):
    '''Handle show of interfaces'''

    global _JSON  # pylint: disable=global-statement
    jsonify = cli.get('json')
    changes = cli.get('changes')
    count = cli.get('count')

    hostname = cli.get('<hostname>') or '*'
    interface = cli.get('<remote-interface>') or '*'

    kind = (
        cli.get('bond') or cli.get('bridge') or cli.get('macvlan') or
        cli.get('vlan') or cli.get('eth') or cli.get('loopback') or
        cli.get('swp') or cli.get('vxlan') or cli.get('vrf') or '*'
    )

    if count:
        count = netq.count_links(hostname=hostname,
                                 ifname=interface, kind=kind)
        if not jsonify:
            print "Matching links count: %d" % count
        else:
            _JSON = {'count': count}
    else:
        if changes:
            headers = ['Node', 'Interface', 'Type', 'State', 'Details',
                       'DbState', 'Last Changed']
            colsizes = [16, 16, 8, 5, 27, 7, 14]
        else:
            headers = ['Node', 'Interface', 'Type', 'State', 'Details',
                       'Last Changed']
            colsizes = [16, 16, 8, 5, 27, 14]

        out = []
        for link in netq.get_links(hostname, interface, kind, changes):
            for entry in link:
                out.append(entry)

        if not out:
            if not jsonify:
                if changes:
                    print "No changes to specified interfaces found"
                else:
                    print 'No matching interface found'

        else:
            if not jsonify:
                print "Matching interface records are:"

            if changes:
                _tabulate(sorted(out, key=lambda x: (x[-1], x[0], x[1]),
                                 reverse=True),
                          headers, jsonify=jsonify,
                          colsizes=colsizes)
            else:
                _tabulate(sorted(out, key=lambda x: (x[0], x[1])),
                          headers, jsonify=jsonify,
                          colsizes=colsizes)

    return _RC_SUCCESS


def cli_show_addresses(cli, netq):
    '''Handle show of IP addresses'''

    global _JSON  # pylint: disable=global-statement

    jsonify = cli.get('json')
    changes = cli.get('changes')
    count = cli.get('count')

    node = cli.get('<hostname>') or '*'
    iface = cli.get('<remote-interface>') or '*'
    vrf = cli.get('<vrf>') or '*'

    if cli.get('ip'):
        ip_address = cli.get('<ipv4>') or cli.get('<ipv4/prefixlen>') or '*'
        is_ipv6 = False
    else:
        ip_address = cli.get('<ipv6>') or cli.get('<ipv6/prefixlen>') or '*'
        is_ipv6 = True

    if count:
        cnt = netq.count_addresses(hostname=node, ifname=iface,
                                   prefix=ip_address, vrf=vrf,
                                   is_ipv6=is_ipv6)

        if not jsonify:
            print "Matching addresses count: %d" % cnt
        else:
            _JSON = {'count': cnt}
    else:
        out = []
        for entry in netq.get_addresses(node, iface, ip_address,
                                        is_ipv6, vrf, changes):
            for addr in entry:
                out.append(addr)

        if not out:
            if not jsonify:
                if changes:
                    print "No changes to interface addresses found"
                else:
                    print 'No matching interface address entry found'
        else:
            if not jsonify:
                print "Matching address records are:"

            if changes:
                headers = ['Address', 'Node', 'Interface', 'VRF',
                           'DbState', 'Last Changed']
                _tabulate(sorted(out, key=lambda x: (x[-1], x[0], x[1],
                                                     x[3]),
                                 reverse=True),
                          headers=headers, jsonify=jsonify)
            else:
                headers = ['Address', 'Node', 'Interface', 'VRF',
                           'Last Changed']
                _tabulate(sorted(out, key=lambda x: (x[0], x[1], x[3])),
                          headers=headers, jsonify=jsonify)

    return _RC_SUCCESS


def cli_show_macs(cli, netq):
    '''Handle showing of MAC entries'''

    global _JSON  # pylint: disable=global-statement

    jsonify = cli.get('json')
    changes = cli.get('changes')
    count = cli.get('count')

    mac = cli.get('<mac>') or '*'
    vlan = cli.get('<0-4096>')

    if vlan is not None:
        vlan = int(vlan)
    else:
        vlan = '*'  # pylint: disable=redefined-variable-type

    hostname = cli.get('<hostname>') or '*'
    oif = cli.get('<interface-nexthop>') or '*'
    origin = cli.get('origin') == 'origin' or '*'
    out = []

    if count:
        cnt = netq.count_macs(mac, vlan, hostname)
        if not jsonify:
            print "Count of matching MAC entries: %d" % cnt
        else:
            _JSON = {'count': cnt}
    else:
        if changes:
            headers = ['Origin', 'MAC', 'VLAN', 'Node Name', 'Egress Port',
                       'DbState', 'Last Changed']
            colsizes = [4, 20, 8, 16, 16, 6, 14]
        else:
            headers = ['Origin', 'MAC', 'VLAN', 'Node Name', 'Egress Port',
                       'Last Changed']
            colsizes = [4, 20, 8, 16, 16, 16]

        for entries in netq.get_macs(mac, vlan, hostname, oif, origin,
                                     changes):
            for entry in entries:
                if changes:
                    out.append(entry)
                else:
                    out.append(entry[0:-2] + [entry[-1]])

        if not out:
            if not jsonify:
                if changes:
                    print "No changes to MAC table found"
                else:
                    print 'No matching MAC table entry found'
        else:
            if not jsonify:
                print "Matching MAC table records are:"

            if changes:
                _tabulate(sorted(out, key=lambda x: (x[-1], x[1], x[0],
                                                     x[2]), reverse=True),
                          headers=headers, jsonify=jsonify,
                          colsizes=colsizes)
            else:
                _tabulate(sorted(out, key=lambda x: (x[1], x[0], x[2]),
                                 reverse=True),
                          headers=headers, jsonify=jsonify,
                          colsizes=colsizes)
    return _RC_SUCCESS


def cli_show_routes(cli, netq):
    '''Handle show of IP routes'''

    global _JSON  # pylint: disable=global-statement

    jsonify = cli.get('json')
    changes = cli.get('changes')
    count = cli.get('count')

    hostname = cli.get('<hostname>') or '*'
    if cli.get('ip'):
        ip_address = cli.get('<ipv4>') or cli.get('<ipv4/prefixlen>') or '*'
        is_ipv6 = False
    else:
        ip_address = cli.get('<ipv6>') or cli.get('<ipv6/prefixlen>') or '*'
        is_ipv6 = True
    if cli.get('origin'):
        origin = True
    else:
        origin = '*'  # pylint: disable=redefined-variable-type
    vrf = cli.get('<vrf>') or '*'

    if count:
        cnt = netq.count_routes(ip_address, hostname, is_ipv6=is_ipv6,
                                origin=origin, vrf=vrf)
        if not jsonify:
            print "Count of matching routes: %s" % cnt
        else:
            _JSON = {'count': cnt}
    else:
        if changes:
            if is_ipv6:
                colsizes = [3, 16, 32, 16, 24, 14, 5]
            else:
                colsizes = [3, 16, 16, 16, 24, 14, 5]
            headers = ['Origin', 'Table', 'IP', 'Node', 'Nexthops',
                       'DbState', 'Last Changed']
        else:
            if is_ipv6:
                colsizes = [3, 16, 32, 16, 26, 16]
            else:
                colsizes = [3, 16, 16, 16, 26, 16]

            headers = ['Origin', 'Table', 'IP', 'Node', 'Nexthops',
                       'Last Changed']

        out = []
        for entries in netq.get_routes(ip_address, hostname, is_ipv6,
                                       origin, vrf, changes):
            for entry in entries:
                out.append(entry)

        if not out:
            if not jsonify:
                if changes:
                    print "No changes to IP routes found"
                else:
                    print 'No matching IP route found'
        else:
            if not jsonify:
                print "Matching IP route records are:"

            if changes:
                _tabulate(sorted(out, key=lambda x: (x[-1], x[1], x[2], x[3],
                                                     x[4]), reverse=True),
                          headers=headers, jsonify=jsonify,
                          colsizes=colsizes)
            else:
                _tabulate(sorted(out, key=lambda x: (x[1], x[2], x[3], x[4])),
                          headers=headers, jsonify=jsonify,
                          colsizes=colsizes)

    return _RC_SUCCESS


def cli_show_ip_neighbors(cli, netq):
    '''Handle show of IP neighbor entries'''

    global _JSON  # pylint: disable=global-statement

    jsonify = cli.get('json')
    changes = cli.get('changes')
    count = cli.get('count')

    if cli.get('ip'):
        ip_address = cli.get('<ipv4>') or '*'
        is_ipv6 = False
    else:
        ip_address = cli.get('<ipv6>') or '*'
        is_ipv6 = True

    mac = cli.get('<mac>') or '*'
    hostname = cli.get('<hostname>') or '*'
    iface = cli.get('<remote-interface>') or '*'
    vrf = cli.get('<vrf>') or '*'

    if count:
        cnt = netq.count_ip_neighbors(hostname=hostname, iface=iface,
                                      ip_address=ip_address, mac=mac,
                                      vrf=vrf, is_ipv6=is_ipv6)
        if not jsonify:
            print "Matching neighbors count: %d" % cnt
        else:
            _JSON = {'count': cnt}
    else:
        out = []
        for entry in netq.get_ip_neighbors(hostname=hostname,
                                           ip_address=ip_address,
                                           ifname=iface, mac=mac,
                                           vrf=vrf, ipv6=is_ipv6,
                                           changes=changes):
            for nbr in entry:
                out.append(nbr)

        if not out:
            if not jsonify:
                if changes:
                    print "No changes to IP neighbor table found"
                else:
                    print 'No matching IP neighbor entry found'

        else:
            if not jsonify:
                print "Matching IP Neighbor records are:"

            if changes:
                headers = ['IP', 'Node', 'Interface', 'MAC', 'VRF',
                           'DbState', 'Last Changed']
                colsizes = [16, 16, 20, 24, 16, 7, 16]
            else:
                headers = ['IP', 'Node', 'Interface', 'MAC', 'VRF',
                           'Last Changed']
                colsizes = [16, 16, 20, 24, 16, 16]

            if changes:
                _tabulate(sorted(out, key=lambda x: (x[-1], x[0], x[2]),
                                 reverse=True),
                          headers=headers, colsizes=colsizes,
                          jsonify=jsonify)
            else:
                _tabulate(sorted(out, key=lambda x: (x[0], x[2]),
                                 reverse=True),
                          headers=headers, colsizes=colsizes,
                          jsonify=jsonify)

    return _RC_SUCCESS


def cli_show_lldp(cli, netq):
    '''Process lldp CLI command'''

    jsonify = cli.get('json')
    node = cli.get('<hostname>') or '*'
    iface = cli.get('<remote-physical-interface>') or '*'

    out = netq.show_neighbors(node, iface)
    if out:
        if not jsonify:
            if node is not None:
                if iface is not None:
                    print 'LLDP peer info for %s:%s' % (node, iface)
                else:
                    print 'LLDP Peer Info for %s' % node
            else:
                print 'LLDP Info for fabric'
        _tabulate(sorted(out),
                  headers=['Node', 'Interface', 'LLDP Peer',
                           'Peer Int', 'Last Changed'],
                  jsonify=jsonify)
    else:
        if not jsonify:
            print 'No LLDP record found'

    return _RC_SUCCESS


def cli_show_bgp_status(cli, netq):
    '''Handle show of BGP status'''

    jsonify = cli.get('json')
    changes = cli.get('changes')

    hostname = cli.get('<hostname>') or '*'
    session = cli.get('<remote-interface>') or cli.get('<ip-session>') or '*'
    asn = cli.get('<number-asn>') or '*'
    vrf = cli.get('<vrf>') or '*'

    out = []
    bad_out = []

    for entries, bad in netq.get_bgp_status_details(hostname, session, asn,
                                                    vrf, changes):
        for entry in entries:
            out.append(entry)
        for bad_entry in bad:
            bad_out.append(bad_entry)

    if changes:
        colsizes = [16, 28, 16, 10, 10, 6, 12, 8, 12]
        headers = ['Node', 'Neighbor', 'VRF', 'ASN', 'Peer ASN',
                   'State', 'PfxRx', 'DbState', 'Last Changed']
    else:
        colsizes = [16, 32, 16, 10, 10, 6, 12, 14]
        headers = ['Node', 'Neighbor', 'VRF', 'ASN', 'Peer ASN',
                   'State', 'PfxRx', 'Last Changed']

    if not out and not bad_out:
        if not jsonify:
            if changes:
                print "No changes to BGP sessions found"
            else:
                print 'No matching BGP sessions found'

    else:
        if not jsonify:
            print "Matching BGP Session records are:"

        if changes:
            _tabulate(sorted([(utils.color_wrap(BgColors.RED, node),
                               utils.color_wrap(BgColors.RED, nbr),
                               utils.color_wrap(BgColors.RED, vrf),
                               utils.color_wrap(BgColors.RED, str(asn)),
                               utils.color_wrap(BgColors.RED, str(pasn)),
                               utils.color_wrap(BgColors.RED, state),
                               utils.color_wrap(BgColors.RED, pfxrx),
                               utils.color_wrap(BgColors.RED, dbstate),
                               last_changed)
                              for (node, nbr, vrf, asn, pasn, state, pfxrx,
                                   dbstate, last_changed) in bad_out] + out,
                             key=itemgetter(8), reverse=True),
                      headers=headers, jsonify=jsonify, colsizes=colsizes)
        else:
            _tabulate(sorted([(utils.color_wrap(BgColors.RED, node),
                               utils.color_wrap(BgColors.RED, nbr),
                               utils.color_wrap(BgColors.RED, vrf),
                               utils.color_wrap(BgColors.RED, str(asn)),
                               utils.color_wrap(BgColors.RED, str(pasn)),
                               utils.color_wrap(BgColors.RED, state),
                               utils.color_wrap(BgColors.RED, pfxrx),
                               utils.color_wrap(BgColors.RED,
                                                utils.pretty_date(uptime)))
                              for (node, nbr, vrf, asn, pasn, state, pfxrx,
                                   uptime) in bad_out] + out,
                             key=lambda x: utils.strip_colors(x[0])),
                      headers=headers, jsonify=jsonify, colsizes=colsizes)

    return _RC_SUCCESS


def cli_show_docker_summary(cli, netq):
    jsonify = bool(cli.get('json'))
    hostname = cli.get('<hostname>') or '*'
    node_id = cli.get('<node-id>') or '*'
    changes = cli.get('changes') or False
    version = cli.get('version>') or '*'
    out = []

    if changes:
        colsizes = [14, 10, 10, 10, 10, 6, 12, 8, 12]
        headers = ['Node', 'Version', 'Installed',
                   'Running', 'Images', 'Swarm Cluster',
                   'Networks', 'Last Changed']
    else:
        colsizes = [14, 10, 10, 10, 10, 6, 12, 8, 12]
        headers = ['Node', 'Version', 'Installed',
                   'Running', 'Images', 'Swarm Cluster',
                   'Networks']

    for docker_host_list in netq.get_docker_host_info(hostname,
                                                      changes=changes,
                                                      version=version):
        for entry in docker_host_list:
            out.append(entry)
    if out:
        if changes:
            _tabulate(sorted(out, key=lambda x: (x[-1], x[0]), reverse=True),
                      headers=headers, jsonify=jsonify)
        else:
            _tabulate(sorted(out, key=lambda x: x[0]),
                      headers=headers, jsonify=jsonify)
    else:
        if not jsonify:
            print 'No Docker Host information found'
    return _RC_SUCCESS


def cli_show_docker_swarm_service(cli, netq):
    jsonify = bool(cli.get('json'))
    hostname = cli.get('<hostname>') or '*'
    service_name = cli.get('<swarm-service-name>') or '*'
    mode_type = cli.get('<mode>') or '*'
    changes = cli.get('changes') or False
    out = []

    if changes:
        headers = ['Service Name', 'Node', 'Mode', 'Replicas', 'Running',
                   'Last changed']
    else:
        headers = ['Service Name', 'Node', 'Mode', 'Replicas', 'Running']

    for docker_swarm_list in netq.get_docker_swarm_service(
            hostname, service_name=service_name, mode=mode_type,
            changes=changes):
        for entry in docker_swarm_list:
            out.append(entry)

    if out:
        if changes:
            _tabulate(sorted(out, key=lambda x: (x[-1], x[0], x[1]),
                            reverse=True), headers=headers, jsonify=jsonify)
        else:
            _tabulate(sorted(out, key=lambda x: (x[0], x[1])),
                      headers=headers, jsonify=jsonify)
    else:
        if not jsonify:
            print 'No Docker Swarm Service information found'
    return _RC_SUCCESS


def cli_show_docker_swarm_service_connectivity(cli, netq):
    jsonify = bool(cli.get('json'))
    hostname = cli.get('<hostname>') or '*'
    service_name = cli.get('<swarm-service-name>') or '*'

    service_info = netq.get_docker_swarm_service_connectivity(hostname=hostname,
                                                              service_name=service_name)

    if jsonify:
        flattened_paths = []
    else:
        flattened_paths = None

    for service, containers in service_info.iteritems():
        if not jsonify:
            print service
            print "   Containers:"
            for container, hosts in containers.iteritems():
                print "    --> " + container
                print "       Hosts:"
                for host, peers in hosts.iteritems():
                    print "       --> " + host
                    print "         Switches:"
                    for peer in peers:
                        print "           --> " + peer
            print
    return _RC_SUCCESS


def cli_show_docker_swarm_service_network(cli, netq):
    jsonify = bool(cli.get('json'))
    hostname = cli.get('<hostname>') or '*'
    service_name = cli.get('<swarm-service-name>') or '*'
    mode_type = cli.get('<mode>') or '*'
    changes = cli.get('changes') or False
    out = []

    if changes:
        headers = ['Service Name', 'Node', 'L4 Ports (proto:hostport:port)',
                   'Virtual IP', 'Network Name', 'Last changed']
    else:
        headers = ['Service Name', 'Node', 'L4 Ports (proto:hostport:port)',
                   'Virtual IP', 'Network Name']

    for docker_swarm_list in netq.get_docker_swarm_service_network(
            hostname, service_name=service_name, mode=mode_type,
            changes=changes):
        for entry in docker_swarm_list:
            out.append(entry)

    if out:
        if changes:
            _tabulate(sorted(out, key=lambda x: (x[-1], x[0], x[1]),
                            reverse=True), headers=headers,
                      jsonify=jsonify)
        else:
            _tabulate(sorted(out, key=lambda x: (x[0], x[1])),
                      headers=headers, jsonify=jsonify)
    else:
        if not jsonify:
            print 'No Docker Swarm Service Network'
    return _RC_SUCCESS


def cli_show_docker_swarm_node(cli, netq):
    hostname = cli.get('<hostname>') or '*'
    cluster_id = cli.get('<cluster-id>') or '*'
    role = cli.get('<role>') or '*'
    jsonify = bool(cli.get('json'))
    changes = cli.get('changes')
    out = []

    if changes:
        headers = ['Host', 'Cluster-name', 'Swarm Node', 'Role',
                   'Docker Version', 'State', 'Plugins', 'Last Changed']
    else:
        headers = ['Host', 'Cluster-name', 'Swarm Node', 'Role',
                   'Docker Version', 'State', 'Plugins']

    for docker_swarm_node_list in netq.get_docker_swarm_node(
            hostname, cluster_id=cluster_id, role=role, changes=changes):
        for entry in docker_swarm_node_list:
            out.append(entry)

    if out:
        if changes:
            _tabulate(sorted(out, key=lambda x: (x[-1], x[0], x[1], x[2]),
                            reverse=True), headers=headers, jsonify=jsonify)
        else:
            _tabulate(sorted(out, key=lambda x: (x[0], x[1], x[2])),
                      headers=headers, jsonify=jsonify)
    else:
        if not jsonify:
            print 'No Docker Swarm Node information found'
    return _RC_SUCCESS


def cli_show_docker_swarm_cluster(cli, netq):
    hostname = cli.get('<hostname>') or '*'
    cluster_id = cli.get('<cluster-id>') or '*'
    cluster_name = cli.get('<cluster-name>') or '*'
    jsonify = bool(cli.get('json'))
    changes = cli.get('changes')
    out = []

    if changes:
        headers = ['Cluster-Name', 'Cluster-id', 'Num Nodes',
                   'Managers', 'Nodes', 'Last changed']
        colsizes = [14, 14, 9, 15, 15, 20]
    else:
        headers = ['Cluster-Name', 'Cluster-id', 'Num Nodes',
                   'Managers', 'Nodes']
        colsizes = [14, 14, 9, 15, 15]

    for docker_swarm_cluster in netq.get_docker_swarm_cluster(
            hostname, cluster_id=cluster_id, cluster_name=cluster_name,
            changes=changes):
        for entry in docker_swarm_cluster:
            out.append(entry)

    if out:
        if changes:
            _tabulate(sorted(out, key=lambda x: (x[-1], x[0]),
                            reverse=True), headers=headers,
                      colsizes=colsizes, jsonify=jsonify)
        else:
            _tabulate(sorted(out, key=lambda x: x[0]), headers=headers,
                      colsizes=colsizes, jsonify=jsonify)
    else:
        if not jsonify:
            print 'No Docker Swarm Cluster information found'
    return _RC_SUCCESS


def cli_show_docker_swarm(cli, netq):
    cluster = cli.get('cluster') or None
    network = cli.get('network') or None
    node = cli.get('node') or None
    retcode = _RC_FAIL

    if node:
        retcode = cli_show_docker_swarm_node(cli, netq)
    elif network:
        retcode = cli_show_docker_swarm_service_network(cli, netq)
    elif cluster:
        retcode = cli_show_docker_swarm_cluster(cli, netq)

    return retcode


def cli_show_docker_container_connectivity(cli, netq):
    hostname = cli.get('<hostname>') or '*'
    ip = cli.get('<ipv4>') or None
    if not ip:
        ip = cli.get('<ipv6>') or '*'
    proto = cli.get('<proto>') or '*'
    port = cli.get('<port>') or '*'
    image = cli.get('<container-image>') or '*'
    container_id = cli.get('<container-id>') or '*'
    name = cli.get('<container-name>') or '*'
    service = cli.get('<swarm-service-name>') or '*'
    network = cli.get('<network-name>') or '*'
    mac = cli.get('<container-mac>') or '*'
    jsonify = bool(cli.get('json'))
    out = []

    headers = ['Name', 'Swarm Service', 'Cont IP', 'Network', 'Node', 'Port',
               'Peer Node', 'Peer Port']
    colsizes = [15, 9, 15, 10, 10, 20, 10, 20 ]

    for container_list in netq.get_docker_container_connectivity(
            hostname, container_ip=ip, container_name=name,
            container_image=image, container_id=container_id,
            service_name=service, network_name=network, mac=mac):
        for entry in container_list:
            out.append(entry)

    if out:
        _tabulate(sorted(out, key=lambda x: (x[0], x[1], x[2])),
                  headers=headers, colsizes=colsizes, jsonify=jsonify)
    else:
        if not jsonify:
            print 'No Docker Container information found'
    return _RC_SUCCESS


def cli_show_docker_container_detail(cli, netq):
    hostname = cli.get('<hostname>') or '*'
    ip = cli.get('<ipv4>') or None
    if not ip:
        ip = cli.get('<ipv6>') or '*'
    proto = cli.get('<proto>') or '*'
    port = cli.get('<port>') or '*'
    image = cli.get('<container-image>') or '*'
    container_id = cli.get('<container-id>') or '*'
    name = cli.get('<container-name>') or '*'
    service = cli.get('<swarm-service-name>') or '*'
    network = cli.get('<network-name>') or '*'
    mac = cli.get('<container-mac>') or '*'
    jsonify = bool(cli.get('json'))
    changes = cli.get('changes') or False
    out = []

    if changes:
        headers = ['Name', 'Node', 'IP', 'IP Masq.', 'Network',
                   'Service Name', 'DBState', 'Last changed']
        colsizes = [20, 10, 17, 8, 14, 14, 8, 15]
    else:
        headers = ['Name', 'Node', 'IP', 'IP Masq.', 'Network',
                   'Service Name', 'Up time']
        colsizes = [20, 10, 17, 8, 14, 15, 15]

    for portmap_list in netq.get_docker_container_detail(hostname, container_ip=ip,
                                                         container_name=name, container_image=image,
                                                         container_id=container_id,
                                                         service_name=service,
                                                         network_name=network,
                                                         mac=mac, changes=changes):
        for entry in portmap_list:
            out.append(entry)

    if out:
        if changes:
            _tabulate(sorted(out, key=lambda x: (x[-1], x[0], x[1], x[2]),
                            reverse=True),
                      headers=headers, colsizes=colsizes, jsonify=jsonify)
        else:
            _tabulate(sorted(out, key=lambda x: (x[0], x[1], x[2])),
                      headers=headers, colsizes=colsizes, jsonify=jsonify)
    else:
        if not jsonify:
            print 'No Docker Container information found'
    return _RC_SUCCESS


def cli_show_docker_container_network(cli, netq):
    hostname = cli.get('<hostname>') or '*'
    ip = cli.get('<ipv4>') or None
    if not ip:
        ip = cli.get('<ipv6>') or '*'
    proto = cli.get('<proto>') or '*'
    port = cli.get('<port>') or '*'
    network_name = cli.get('<network-name>') or '*'
    subnet = cli.get('<ipv4/prefixlen>') or '*'
    jsonify = bool(cli.get('json'))
    changes = cli.get('changes')
    exposed = cli.get('exposed')
    out = []

    if changes:
        headers = ['Container Name', 'Node', 'Proto', 'Port', 'IP',
                   'Network', 'Host IP', 'Host Port','DBState', 'Last Changed']
        colsizes = [20, 10, 6, 8, 17, 14, 21, 12, 8, 12]
    else:
        headers = ['Container Name', 'Node', 'Proto', 'Port', 'Cont IP',
                   'Network', 'Host IP', 'Host Port']
        colsizes = [20, 10, 6, 8, 17, 14, 21, 12]

    for portmap_list in netq.get_docker_container_port(hostname, container_ip=ip,
                                                       proto=proto, port=port,
                                                       subnet=subnet,
                                                       network_name=network_name,
                                                       changes=changes):
        for entry in portmap_list:
            out.append(entry)

    if out:
        if changes:
            _tabulate(sorted(out, key=lambda x: (x[-1], x[1], x[0]),
                            reverse=True), headers=headers,
                      colsizes=colsizes, jsonify=jsonify)
        else:
            _tabulate(sorted(out, key=lambda x: (x[1], x[0])), headers=headers,
                      colsizes=colsizes, jsonify=jsonify)
    else:
        if not jsonify:
            print 'No Docker container found'
    return _RC_SUCCESS


def cli_show_docker_container_adjacent(cli, netq):
    hostname = cli.get('<hostname>') or None
    interface = cli.get('<remote-physical-interface>') or '*'
    service_name = cli.get('<swarm-service-name>') or '*'
    container_name = cli.get('<container-name>') or '*'
    ip = cli.get('<ipv4>') or None
    if not ip:
        ip = cli.get('<ipv6>') or '*'
    jsonify = bool(cli.get('json'))
    retcode = _RC_FAIL
    out = []

    if not hostname:
        if not jsonify:
            print 'No Docker container found'
        return retcode

    headers = [ 'Interface', 'Peer Node', 'Peer Interface',
                'Container Name', 'IP', 'Network',
                'Service Name']
    colsizes = [20, 10, 20, 20, 20, 10, 15]

    for container_list in netq.get_docker_container_adjacent(hostname, interface=interface,
                                                             service_name=service_name,
                                                             container_name=container_name,
                                                             container_ip=ip):
        for entry in container_list:
            out.append(entry)

    if out:
        _tabulate(sorted(out, key=lambda x: (x[1], x[0])), headers=headers,
                  colsizes=colsizes, jsonify=jsonify)
    else:
        if not jsonify:
            print 'No Docker container found'
    return _RC_SUCCESS


def cli_show_docker_container(cli, netq):
    portmap = cli.get('portmap') or None
    proto = cli.get('<proto>') or None
    adjacent = cli.get('adjacent') or None
    connectivity = cli.get('connectivity') or None

    if adjacent:
        retcode = cli_show_docker_container_adjacent(cli, netq)
    elif connectivity:
        retcode = cli_show_docker_container_connectivity(cli, netq)
    elif portmap or proto:
        retcode = cli_show_docker_container_network(cli, netq)
    else:
        retcode = cli_show_docker_container_detail(cli, netq)
    return retcode


def cli_show_docker_network_detail(cli, netq):
    hostname = cli.get('<hostname>') or '*'
    network_name = cli.get('<network-name>') or '*'
    network_driver = cli.get('<network-driver>') or '*'
    subnet = cli.get('<ipv4/prefixlen>') or '*'
    jsonify = bool(cli.get('json'))
    changes = cli.get('changes') or False
    out = []

    if changes:
        headers = ['Network Name', 'Node', 'Driver', 'subnet',
                   'gateway', 'ipv6', 'ip masq.', 'encrypted',
                   'Containers', 'Last changed']
        colsizes = [16, 10, 9, 15, 15, 9, 6, 9, 25, 12]
    else:
        headers = ['Network Name', 'Node', 'Driver', 'subnet',
                   'gateway', 'ipv6', 'ip masq.', 'encrypted',
                   'Containers']
        colsizes = [16, 10, 9, 15, 15, 9, 6, 9, 25]

    for network_list in netq.get_docker_network_detail(
            hostname, network_name=network_name, subnet=subnet,
            network_driver=network_driver, changes=changes):
        for entry in network_list:
            out.append(entry)

    if out:
        if changes:
            _tabulate(sorted(out, key=lambda x: (x[-1], x[0], x[1]),
                            reverse=True), headers=headers,
                      colsizes=colsizes, jsonify=jsonify)
        else:
            _tabulate(sorted(out, key=lambda x: (x[0], x[1])), headers=headers,
                      colsizes=colsizes, jsonify=jsonify)
    else:
        if not jsonify:
            print 'Docker Network information not found'
    return _RC_SUCCESS


def cli_show_docker_network_brief(cli, netq):
    hostname = cli.get('<hostname>') or '*'
    network_name = cli.get('<network-name>') or '*'
    network_driver = cli.get('<network-driver>') or '*'
    subnet = cli.get('<ipv4/prefixlen>') or '*'
    jsonify = bool(cli.get('json'))
    changes = cli.get('changes')
    out = []

    if changes:
        headers = ['Network Name', 'Node', 'subnet', 'gateway', 'ipv6',
                   'ip masq.', 'encrypted', 'Last changed']
        colsizes = [16, 10, 15, 15, 9, 6, 9, 12]
    else:
        headers = ['Network Name', 'Node', 'subnet', 'gateway', 'ipv6',
                   'ip masq.', 'encrypted']
        colsizes = [16, 10, 15, 15, 9, 6, 9]

    for network_list in netq.get_docker_network_brief(
            hostname, network_name=network_name, subnet=subnet,
            network_driver=network_driver, changes=changes):
        for entry in network_list:
            out.append(entry)

    if out:
        if changes:
            _tabulate(sorted(out, key=lambda x: (x[-1], x[0], x[1]),
                             reverse=True), headers=headers,
                      colsizes=colsizes, jsonify=jsonify)
        else:
            _tabulate(sorted(out, key=lambda x: (x[0], x[1])), headers=headers,
                      colsizes=colsizes, jsonify=jsonify)
    else:
        if not jsonify:
            print 'Docker Network information not found'
    return _RC_SUCCESS


def cli_show_docker_network(cli, netq):
    brief = cli.get('<brief>') or None
    network_name = cli.get('<network-name>') or None
    network_driver = cli.get('<network-driver>') or None
    subnet = cli.get('<ipv4/prefixlen>') or None
    retcode = _RC_FAIL

    if (network_name or network_driver or subnet) and not brief:
        retcode = cli_show_docker_network_detail(cli, netq)
    else:
        retcode = cli_show_docker_network_brief(cli, netq)

    return retcode

def cli_show_docker_service_impact(cli, netq):
    '''show docker service impact'''
    global _JSON  # pylint: disable=global-statement
    hostname = cli.get('<hostname>') or '*'
    service_name = cli.get('<swarm-service-name>') or '*'
    jsonify = cli.get('json')
    headers = ['Service Name', 'Desired count', 'Running count ',
               'Total loss count', 'Partial loss count']
    colsizes = [16, 9, 10, 10, 10]
    rc = _RC_SUCCESS

    out, counts = netq.show_docker_service_impact(hostname, service_name=service_name)

    if not out:
        if not jsonify:
            print 'No Docker Service impact found'
        else:
            out_json = {}
            out_json['summary'] = {
                'checkedNodeCount': len(counts.get('checkedNodes')),
                'failedNodeCount': len(counts.get('failedNodes'))
            }
            _JSON = out_json
    else:
        out_list = []
        for entry in out:
            if entry[3] != 0:
                out_list.append((utils.color_wrap(BgColors.RED,
                                                  str(entry[0])),
                                 entry[1], entry[2],
                                 utils.color_wrap(BgColors.RED,
                                                  str(entry[3])),
                                 entry[4]))
            else:
                out_list.append((entry[0], entry[1], entry[2],
                                 entry[3], entry[4]))

        if out_list:
            _tabulate(sorted(out_list, key=lambda x: x[0]),
                      headers=headers, colsizes=colsizes,
                      jsonify=jsonify)

        if jsonify:
            out_json = {}
            out_list = []
            for entry in out:
                out_list.append({'service_name': entry[0],
                                 'desired_count': entry[1],
                                 'running_count': entry[2],
                                 'total_conn_loss_count': entry[3],
                                 'partial_conn_loss_count': entry[4]})
            out_json['description'] = json.dumps(out_list)
            out_json['summary'] = {
                'checkedNodeCount': len(counts.get('checkedNodes')),
                'failedNodeCount': len(counts.get('failedNodes'))
            }
            _JSON = out_json

    return rc



def cli_show_docker(cli, netq):
    host = cli.get('summary') or None
    swarm = cli.get('swarm') or None
    container = cli.get('container') or None
    network = cli.get('network') or None
    service = cli.get('service') or None
    connectivity = cli.get('connectivity') or None
    impact = cli.get('impact') or None
    retcode = _RC_FAIL


    if host:
        retcode = cli_show_docker_summary(cli, netq)
    elif swarm:
        retcode = cli_show_docker_swarm(cli, netq)
    elif container:
        retcode = cli_show_docker_container(cli, netq)
    elif network:
        retcode = cli_show_docker_network(cli, netq)
    elif connectivity:
        retcode = cli_show_docker_swarm_service_connectivity(cli, netq)
    elif service:
        retcode = cli_show_docker_swarm_service(cli, netq)
    elif impact:
        retcode = cli_show_docker_service_impact(cli, netq)
    else:
        retcode = _RC_FAIL

    return retcode


def cli_show_lnv_status(cli, netq):
    '''Handle show of LNV status'''

    jsonify = cli.get('json')
    changes = cli.get('changes')

    hostname = cli.get('<hostname>') or '*'
    out = []
    bad_out = []

    for entries, bad in netq.get_lnv_status_details(hostname, changes):
        for entry in entries:
            out.append(entry)

        for bad_entry in bad:
            bad_out.append(bad_entry)

    if changes:
        colsizes = [16, 4, 8, 5, 6, 6, 3, 14]
        headers = ['Node', 'Role', 'ReplMode', 'State', '#Peers', '#Vnis',
                   'DbState', 'Last Changed']
    else:
        colsizes = [16, 4, 8, 5, 6, 6, 14]
        headers = ['Node', 'Role', 'ReplMode', 'State', '#Peers', '#Vnis',
                   'Last Changed']

    if not out:
        if not jsonify:
            if changes:
                print "No changes to LNV session found"
            else:
                print 'No matching LNV session entries found'

    else:
        if not jsonify:
            print "Matching LNV session records are:"

        if changes:
            _tabulate(sorted([(utils.color_wrap(BgColors.RED, node),
                               utils.color_wrap(BgColors.RED, role),
                               utils.color_wrap(BgColors.RED, replmode),
                               utils.color_wrap(BgColors.RED, state),
                               utils.color_wrap(BgColors.RED, str(tpeers)),
                               utils.color_wrap(BgColors.RED, str(tvnis)),
                               utils.color_wrap(BgColors.RED, dbs),
                               timestamp)
                              for (node, role, replmode, state, tpeers, tvnis,
                                   dbs, timestamp) in bad_out] + out,
                             key=itemgetter(7), reverse=True),
                      headers, colsizes=colsizes, jsonify=jsonify)
        else:
            _tabulate(sorted([(utils.color_wrap(BgColors.RED, node),
                               utils.color_wrap(BgColors.RED, role),
                               utils.color_wrap(BgColors.RED, replmode),
                               utils.color_wrap(BgColors.RED, state),
                               utils.color_wrap(BgColors.RED, str(tpeers)),
                               utils.color_wrap(BgColors.RED, str(tvnis)),
                               ts)
                              for (node, role, replmode, state, tpeers, tvnis,
                                   ts) in bad_out] + out,
                             key=lambda x: utils.strip_colors(x[0])),
                      headers, colsizes=colsizes, jsonify=jsonify)

    return _RC_SUCCESS


def cli_show_clag_status(cli, netq):
    '''Handle show of CLAG status'''

    jsonify = cli.get('json')
    changes = cli.get('changes')

    hostname = cli.get('<hostname>') or '*'
    out = []
    bad_out = []

    for entries, bad in netq.get_clag_status_details(hostname, changes):
        for entry in entries:
            out.append(entry)

        for bad_entry in bad:
            bad_out.append(bad_entry)

    if changes:
        colsizes = [16, 16, 17, 4, 6, 5, 6, 3, 14]
        headers = ['Node', 'Peer', 'SysMac', 'State', 'Backup',
                   '#Bonds', '#Dual', 'DbState', 'Last Changed']
    else:
        colsizes = [16, 16, 17, 4, 6, 6, 5, 14]
        headers = ['Node', 'Peer', 'SysMac', 'State', 'Backup',
                   '#Links', '#Dual', 'Last Changed']

    if not out and not bad_out:
        if not jsonify:
            if changes:
                print "No changes to CLAG session found"
            else:
                print 'No matching CLAG session entries found'

    else:
        if not jsonify:
            print "Matching CLAG session records are:"

        if changes:
            _tabulate(sorted([(utils.color_wrap(BgColors.RED, node),
                               utils.color_wrap(BgColors.RED, peer),
                               utils.color_wrap(BgColors.RED, sysmac),
                               utils.color_wrap(BgColors.RED, state),
                               utils.color_wrap(BgColors.RED, backup),
                               utils.color_wrap(BgColors.RED, str(tbonds)),
                               utils.color_wrap(BgColors.RED, str(dbonds)),
                               utils.color_wrap(BgColors.RED, dbs),
                               timestamp)
                              for (node, peer, sysmac, state, backup,
                                   tbonds, dbonds, dbs,
                                   timestamp) in bad_out] + out,
                             key=itemgetter(8), reverse=True),
                      headers, colsizes=colsizes, jsonify=jsonify)
        else:
            _tabulate(sorted([(utils.color_wrap(BgColors.RED, node),
                               utils.color_wrap(BgColors.RED, peer),
                               utils.color_wrap(BgColors.RED, sysmac),
                               utils.color_wrap(BgColors.RED, state),
                               utils.color_wrap(BgColors.RED, backup),
                               utils.color_wrap(BgColors.RED, str(tbonds)),
                               utils.color_wrap(BgColors.RED, str(dbonds)),
                               ts)
                              for (node, peer, sysmac, state, backup,
                                   tbonds, dbonds, ts) in bad_out] + out,
                             key=lambda x: utils.strip_colors(x[0])),
                      headers, colsizes=colsizes, jsonify=jsonify)

    return _RC_SUCCESS


def cli_show_stp_topology(cli, netq):
    '''Process STP CLI'''

    global _JSON  # pylint: disable=global-statement

    jsonify = cli.get('json')

    hostname = cli.get('<hostname>') or '*'

    if jsonify:
        flattened_paths = []
    else:
        flattened_paths = None

    error = []
    stpmap = netq.get_stp_topo(hostname, True, error)
    if error and not stpmap:
        if not jsonify:
            print utils.color_wrap(BgColors.RED, '\n'.join(error))
        else:
            _JSON = {'error': '\n'.join(error)}
        return _RC_FAIL
    if stpmap:
        utils.prtree(stpmap[0], 0, True, [], flattened_paths)
        if len(stpmap) == 2:
            if jsonify:
                peerpaths = flattened_paths
                flattened_paths = []
            else:
                print
            utils.prtree(stpmap[1][0], 0, True, [], flattened_paths)
            if jsonify:
                flattened_paths += peerpaths

        if jsonify:
            _JSON = flattened_paths
        else:
            print

    return _RC_SUCCESS

def cli_show_services(cli, netq):
    '''Process CLI for services'''

    jsonify = cli.get('json')
    changes = cli.get('changes')
    count = cli.get('count')

    hostname = cli.get('<hostname>') or '*'
    svcname = cli.get('<service-name>') or '*'
    vrf = cli.get('<vrf>') or '*'
    active = cli.get('active') == 'active' or '*'
    monitored = cli.get('monitored') == 'monitored' or '*'
    status = (cli.get('ok') or cli.get('warning') or cli.get('error') or
              cli.get('fail') or '*')
    out = []

    if count:
        service_count = netq.get_services_count(hostname, svcname, vrf,
                                                active, status, monitored)
        if not jsonify:
            print "Matching Service Count: %d" % service_count
        else:
            _JSON = {'count': service_count}
    else:
        for entry in netq.get_services_status(hostname, svcname, vrf,
                                              active, status,
                                              monitored, changes):
            for svc in entry:
                out.append(svc)

        if not out:
            if not jsonify:
                if changes:
                    print "No changes to services found"
                else:
                    print 'No matching service entries found'

        else:
            if not jsonify:
                print "Matching services records are:"

            if changes:
                _tabulate(sorted(out, key=lambda x: (x[-1], x[0], x[1]),
                                 reverse=True),
                          headers=['Node', 'Service', 'PID', 'VRF', 'Enabled',
                                   'Active', 'Monitored', 'Status', 'Up Time',
                                   'DbState', 'Last Changed'],
                          jsonify=jsonify)
            else:
                _tabulate(sorted(out, key=lambda x: (x[0], x[1])),
                          headers=['Node', 'Service', 'PID', 'VRF', 'Enabled',
                                   'Active', 'Monitored', 'Status', 'Up Time',
                                   'Last Changed'],
                          jsonify=jsonify)

    return _RC_SUCCESS


def cli_show_inventory(cli, netq):
    '''Process inventory CLI'''

    jsonify = cli.get('json')

    hostname = cli.get('<hostname>') or '*'

    if cli.get('os'):
        os_version_id = cli.get('<os-version>') or None
        os_name = cli.get('<os-name>') or None
        search_str = os_name or os_version_id or None
        out = netq.get_os_status(hostname, search_str)
        _tabulate(sorted(out), ['Node', 'OS Name', 'OS Version'],
                  jsonify=jsonify)
    elif cli.get('board'):
        vendor = cli.get('<board-vendor>') or None
        model = cli.get('<board-model>') or None
        search_str = vendor or model or None
        out = netq.get_board_status(hostname, search_str)
        _tabulate(sorted(out), ['Node', 'Vendor', 'Model', 'Base MAC',
                                'Serial No', 'Part No', 'Rev', 'Mfg Date'],
                  jsonify=jsonify)
    elif cli.get('cpu'):
        arch = cli.get('<cpu-arch>') or None
        model = cli.get('<cpu-model>') or None
        search_str = arch or model or None
        out = netq.get_cpu_status(hostname, search_str)
        _tabulate(
            sorted(out),
            ['Node', 'Arch', 'Model', 'Freq', 'Cores', 'Memory'],
            jsonify=jsonify)
    elif cli.get('asic'):
        vendor = cli.get('<asic-vendor>') or None
        model = cli.get('<asic-model>') or None
        model_id = cli.get('<asic-model-id>') or None
        search_str = vendor or model or model_id or None
        out = netq.get_asic_status(hostname, search_str)
        _tabulate(
            sorted(out),
            ['Node', 'Vendor', 'Model', 'Model ID', 'Core B/W', 'Ports'],
            jsonify=jsonify)
    elif cli.get('license'):
        jsonify = cli.get('json')
        changes = cli.get('changes')
        name = Platform.CL if cli.get('cumulus') else '*'
        hostname = cli.get('<hostname>') or '*'
        headers = ['Node', 'Name', 'State', 'License', 'Last Changed']
        out = []
        bad_out = []
        for entries, bad in netq.get_license_status(hostname, name, changes):
            for entry in entries:
                out.append(entry)
            for bad_entry in bad:
                bad_out.append(bad_entry)

        if not out and not bad_out:
            if not jsonify:
                if changes:
                    print('No changes to licenses found')
                else:
                    print('No matching licenses found')
        else:
            _tabulate(sorted(
                [(utils.color_wrap(BgColors.RED, node),
                  utils.color_wrap(BgColors.RED, name),
                  utils.color_wrap(BgColors.RED, state),
                  utils.color_wrap(BgColors.RED, lic), last_changed)
                 for (node, name, state, lic, last_changed) in
                 bad_out] + out,
                key=itemgetter(2), reverse=True), headers=headers,
                jsonify=jsonify)

        return _RC_SUCCESS

    elif cli.get('disk'):
        name = cli.get('<disk-name>') or None
        size = cli.get('<disk-size>') or None
        transport = cli.get('<disk-transport>') or None
        vendor = cli.get('<disk-vendor>') or None
        search_str = name or size or transport or vendor or None
        out = netq.get_disk_status(hostname, search_str)
        _tabulate(
            sorted(out),
            ['Node', 'Name', 'Type', 'Transport', 'Size', 'Vendor',
             'Model'],
            jsonify=jsonify)
    elif cli.get('memory'):
        m_type = cli.get('<memory-type>') or None
        size = cli.get('<memory-size>') or None
        vendor = cli.get('<memory-vendor>') or None
        search_str = m_type or size or vendor or None
        out = netq.get_memory_status(hostname, search_str)
        _tabulate(
            sorted(out),
            ['Node', 'Name', 'Type', 'Size', 'Speed', 'Vendor',
             'Serial No'],
            jsonify=jsonify)
    else:
        out = []
        search_str = cli.get('<inventory-search>') or None
        if hostname is not '*':
            out = netq.get_inventory_status(hostname)
            if len(out):
                _tabulate(sorted(out),
                          ['Node', 'Switch', 'OS', 'CPU', 'ASIC', 'Ports'],
                          jsonify=jsonify)
                return _RC_SUCCESS
            search_str = hostname
            hostname = '*'
        if not search_str:
            out = netq.get_inventory_status(hostname)
            if len(out):
                _tabulate(sorted(out),
                          ['Node', 'Switch', 'OS', 'CPU', 'ASIC', 'Ports'],
                          jsonify=jsonify)
                return _RC_SUCCESS
        out = netq.get_os_status(hostname, search_str)
        if len(out):
            _tabulate(sorted(out), ['Node', 'OS Name', 'OS Version'],
                      jsonify=jsonify)
            return _RC_SUCCESS
        out = netq.get_cpu_status(hostname, search_str)
        if len(out):
            _tabulate(
                sorted(out),
                ['Node', 'Arch', 'Model', 'Freq', 'Cores', 'Memory'],
                jsonify=jsonify)
            return _RC_SUCCESS
        out = netq.get_asic_status(hostname, search_str)
        if len(out):
            _tabulate(
                sorted(out),
                ['Node', 'Vendor', 'Model', 'Model ID', 'Core B/W', 'Ports'],
                jsonify=jsonify)
            return _RC_SUCCESS
        out = netq.get_board_status(hostname, search_str)
        if len(out):
            _tabulate(sorted(out),
                      ['Node', 'Vendor', 'Model', 'Base MAC',
                       'Serial No', 'Part No', 'Rev', 'Mfg Date'],
                      jsonify=jsonify)
            return _RC_SUCCESS
        out = netq.get_disk_status(hostname, search_str)
        if len(out):
            _tabulate(
                sorted(out),
                ['Node', 'Name', 'Type', 'Transport', 'Size', 'Vendor',
                 'Model'],
                jsonify=jsonify)
            return _RC_SUCCESS
        out = netq.get_memory_status(hostname, search_str)
        if len(out):
            _tabulate(
                sorted(out),
                ['Node', 'Name', 'Type', 'Size', 'Speed', 'Vendor',
                 'Serial No'],
                jsonify=jsonify)
            return _RC_SUCCESS

    return _RC_SUCCESS


def cli_show_agent_status(cli, netq):
    '''Show/check agent status'''

    jsonify = cli.get('json')
    changes = cli.get('changes')
    hostname = cli.get('<hostname>') or '*'

    if changes:
        out = []
        restarts = 0
        for key, grouper in itertools.groupby(
                netq.show_agent_changes(hostname=hostname),
                key=itemgetter(0)
        ):
            items = sorted(grouper, key=itemgetter(2))
            if len(items) > 1:
                restarts = len(items) - 1
                color = BgColors.RED
            else:
                restarts = 0
                color = BgColors.GREEN

            if not int(items[-1][-1]):
                cur_color = BgColors.RED
            else:
                cur_color = BgColors.GREEN

            if hostname != '*':
                # Provide detailed output for this scenario
                first = True
                for item in items:
                    ts1 = item[2]
                    if not jsonify:
                        ts1 = datetime.utcfromtimestamp(int(ts1))
                        uptime = datetime.utcfromtimestamp(int(item[1]))
                    else:
                        uptime = item[1]
                    if first:
                        out.append([utils.color_wrap(cur_color, key),
                                    utils.color_wrap(cur_color, str(uptime)),
                                    utils.color_wrap(cur_color, str(ts1))])
                        first = False
                    else:
                        out.append([key, str(uptime), str(ts1)])

            else:
                ts1 = items[0][2]
                ts2 = items[-1][2]
                if not jsonify:
                    ts1 = datetime.utcfromtimestamp(int(ts1))
                    ts2 = datetime.utcfromtimestamp(int(ts2))
                out.append([utils.color_wrap(cur_color, key),
                            str(ts1),
                            utils.color_wrap(cur_color, str(ts2)),
                            utils.color_wrap(color, '%d' % restarts)])

        if hostname == '*':
            _tabulate(out,
                      headers=['Node', 'First Connect time',
                               'Last Connect Time',
                               'NumRestarts'], jsonify=jsonify)
        else:
            if not jsonify and restarts:
                print utils.color_wrap(BgColors.RED,
                                       'Number of Restarts: %d' %
                                       restarts)
            _tabulate(out, headers=['Node', 'Sys Bootup Time', 'Connect Time'],
                      jsonify=jsonify)
    else:
        out = netq.show_status(hostname)
        new_out = []
        for nodename, status, uptime, ts1 in out:
            if status == AgentHealth.ROTTEN:
                new_status = utils.color_wrap(BgColors.RED, 'Rotten')
                if jsonify:
                    ts1 = 0.0
                    uptime = 0.0
                else:
                    ts1 = '-'
                    uptime = '-'
            else:
                new_status = utils.color_wrap(BgColors.GREEN, 'Fresh')
            bisect.insort(new_out, (nodename, new_status, uptime, ts1))

        _tabulate(sorted(new_out),
                  headers=['Node', 'Status', 'Sys Uptime', 'Agent Uptime'],
                  jsonify=jsonify)


def cli_show_sensors(cli, netq):
    """Handle sensors show CLI"""
    jsonify = cli.get('json')
    changes = cli.get('changes')
    hostname = cli.get('<hostname>') or '*'
    out = []
    temp = cli.get('temp')
    fan = cli.get('fan')
    psu = cli.get('psu')

    if temp:
        name = cli.get('<temp-name>') or '*'
        for entries in netq.get_temp_status(hostname, name, changes):
            for entry in entries:
                out.append(entry)
        if changes:
            headers = ['Node', 'Name', 'Description', 'State', 'Temp',
                       'Critical', 'Max', 'Min', 'Message', 'DbState',
                       'Last Changed']
        else:
            headers = ['Node', 'Name', 'Description', 'State', 'Temp',
                       'Critical', 'Max', 'Min', 'Message']
    elif fan:
        name = cli.get('<fan-name>') or '*'
        for entries in netq.get_fan_status(hostname, name, changes):
            for entry in entries:
                out.append(entry)
        if changes:
            headers = ['Node', 'Name', 'Description', 'State', 'Speed', 'Max',
                       'Min', 'Message', 'DbState', 'Last Changed']
        else:
            headers = ['Node', 'Name', 'Description', 'State', 'Speed', 'Max',
                       'Min', 'Message']
    elif psu:
        name = cli.get('<psu-name>') or '*'
        for entries in netq.get_psu_status(hostname, name, changes):
            for entry in entries:
                out.append(entry)
        if changes:
            headers = ['Node', 'Name', 'State', 'Message', 'DbState',
                       'Last Changed']
        else:
            headers = ['Node', 'Name', 'State', 'Message']
    else:
        for entries in netq.get_temp_status(hostname, '*', changes):
            for x in entries:
                if changes:
                    out.append([x[0], x[1], x[2], x[3],
                                x[-3], x[-2], x[-1]])
                else:
                    out.append([x[0], x[1], x[2], x[3], x[-1]])
        for entries in netq.get_fan_status(hostname, '*', changes):
            for x in entries:
                if changes:
                    out.append([x[0], x[1], x[2], x[3],
                                x[-3], x[-2], x[-1]])
                else:
                    out.append([x[0], x[1], x[2], x[3], x[-1]])
        for entries in netq.get_psu_status(hostname, '*', changes):
            for x in entries:
                if changes:
                    out.append([x[0], x[1], 'N/A', x[2],
                                x[3], x[4], x[5]])
                else:
                    out.append([x[0], x[1], 'N/A', x[2], x[3]])
        if changes:
            headers = ['Node', 'Name', 'Description', 'State', 'Message',
                       'DbState', 'Last Changed']
        else:
            headers = ['Node', 'Name', 'Description', 'State', 'Message']

    key = lambda x: (x[0], x[1])
    reverse = False
    if changes:
        key = lambda x: (x[-1], x[0], x[1])
        reverse = True

    _tabulate(sorted(out, key=key, reverse=reverse), headers=headers,
              jsonify=jsonify)
    return _RC_SUCCESS


def cli_show_changes(cli, netq):
    '''Process changes CLI option'''

    global _JSON  # pylint: disable=global-statement

    jsonify = cli.get('json')
    mac = cli.get('<mac>') or '*'
    ip_address = (
        cli.get('<ipv4>') or cli.get('<ipv4/prefixlen>') or
        cli.get('<ipv6>') or cli.get('<ipv6/prefixlen>') or '*'
    )

    out_json = []

    if ip_address == '*' and mac == '*':
        rc = cli_show_links(cli, netq)
        if jsonify and _JSON:
            out_json.append([{'links': _JSON[0]}])
            _JSON = None

    if mac == '*':
        rc = cli_show_addresses(cli, netq)
        if jsonify and _JSON:
            out_json.append([{'addresses': _JSON[0]}])
            _JSON = None

    if ip_address == '*':
        rc = cli_show_macs(cli, netq)
        if jsonify and _JSON:
            out_json.append([{'macs': _JSON[0]}])
            _JSON = None

    if mac == '*':
        rc = cli_show_routes(cli, netq)
        if jsonify and _JSON:
            out_json.append([{'routes': _JSON[0]}])
            _JSON = None

    rc = cli_show_ip_neighbors(cli, netq)
    if jsonify and _JSON:
        out_json.append([{'neighbors': _JSON[0]}])
        _JSON = None

    if ip_address == '*' and mac == '*':
        rc = cli_show_bgp_status(cli, netq)
        if jsonify and _JSON:
            out_json.append([{'bgpd': _JSON[0]}])
            _JSON = None

        rc = cli_show_clag_status(cli, netq)
        if jsonify and _JSON:
            out_json.append([{'clagd': _JSON[0]}])
            _JSON = None

        rc = cli_show_lnv_status(cli, netq)
        if jsonify and _JSON:
            out_json.append([{'lnv': _JSON[0]}])
            _JSON = None

    # JSON Processing at the end
    if out_json:
        _JSON = out_json


def process_daemon(name, cli):
    '''Common systemctl commands function for daemons'''
    cmds = ['start', 'stop', 'status', 'restart']
    cmd_issued = None
    systemctl = '/bin/systemctl'
    fnull = open(os.devnull, 'w')
    rc = _RC_SUCCESS
    vrf = None

    for cmd in cmds:
        cmd_issued = cli.get(cmd)
        if cmd_issued:
            break

    conf_dict = utils.get_yaml_config()
    if conf_dict and ConfKw.BACKEND in conf_dict:
        vrf = conf_dict[ConfKw.BACKEND].get(ConfKw.VRF, None)
        if vrf == 'default':
            vrf = None

    if vrf:
        name += '@%s' % vrf

    if cmd_issued == 'start':
        if os.getuid() != 0:
            print 'Please run this command as root'
            rc = _RC_FAIL
        try:
            subprocess.check_output([systemctl, 'enable', name], stderr=fnull)
            subprocess.check_output([systemctl, 'reset-failed', name],
                                    stderr=fnull)
            subprocess.check_output([systemctl, 'start', name], stderr=fnull)
            print(utils.color_wrap(BgColors.GREEN, 'Success!'))
        except subprocess.CalledProcessError:
            print 'Failed to start %s' % name
            rc = _RC_FAIL
    elif cmd_issued == 'stop':
        if os.getuid() != 0:
            print 'Please run this command as root'
            rc = _RC_FAIL
        try:
            subprocess.check_output([systemctl, 'stop', name], stderr=fnull)
            subprocess.check_output([systemctl, 'disable', name], stderr=fnull)
            print(utils.color_wrap(BgColors.GREEN, 'Success!'))
        except subprocess.CalledProcessError:
            print 'Failed to stop %s' % name
            rc = _RC_FAIL
    elif cmd_issued == 'status':
        try:
            subprocess.check_output([systemctl, 'status', name], stderr=fnull)
            print(utils.color_wrap(BgColors.GREEN, 'Running...'))
        except subprocess.CalledProcessError:
            print 'Stopped...'
            rc = _RC_FAIL
    elif cmd_issued == 'restart':
        try:
            subprocess.check_output([systemctl, 'enable', name], stderr=fnull)
            subprocess.check_output([systemctl, 'reset-failed', name],
                                    stderr=fnull)
            subprocess.check_output([systemctl, 'restart', name], stderr=fnull)
            print(utils.color_wrap(BgColors.GREEN, 'Success!'))
        except subprocess.CalledProcessError:
            print 'Failed to restart %s' % name
            rc = _RC_FAIL
    else:
        print 'Invalid command %s for %s' % (cmd_issued, name)
        rc = _RC_FAIL

    return rc


def cli_process_agent(cli, netq):
    return process_daemon('netq-agent', cli)


def cli_process_notifier(cli, netq):
    return process_daemon('netq-notifier', cli)


def cli_resolve(cli, netq):
    '''Convert numbers into names from the DB'''

    vrf = cli.get('<vrf>') or '*'
    vlan = cli.get('<vlan>') or '*'
    ctx = {}
    pmtu_error = ''
    if not sys.stdin.isatty():
        for line in sys.stdin:
            line = netq.resolve_output(line, vrf, vlan, ctx)
            print line,
            if 'pmtu_error' in ctx and not pmtu_error:
                pmtu_error = ctx['pmtu_error']

        if (ctx['mtu'] and ctx['last_mtu'] and
                ctx['mtu'] != ctx['last_mtu']):
            if 'traceroute' in ctx:
                print utils.color_wrap(
                    BgColors.RED,
                    'Starting & ending links have differing MTU')
        if pmtu_error:
            print(
                utils.color_wrap(
                    BgColors.RED,
                    '* - Links with mismatched MTU from origin %s' %
                    ctx['mtu_string'])
            )
        if 'traceroute' in ctx:
            if ctx['mtu'] == 0:
                print('Path MTU couldn\'t be determined')
            else:
                print('Path MTU is %d' % ctx['mtu'])

    return _RC_SUCCESS


def cli_check_vlan(cli, netq):
    '''Perform VLAN consistency check'''

    global _JSON  # pylint: disable=global-statement

    jsonify = cli.get('json')

    hostname = cli.get('<hostname>') or '*'
    ifname = cli.get('<remote-interface>')  # we want this to be not '*'
    unverified = cli.get('unverified')
    retcode = _RC_SUCCESS

    unverified_lst, counts, output = (
        netq.check_vlan_consistency(hostname, ifname))

    out_json = {'failedNodes': [],
                'unverifiedLinks': [],
                'summary': counts}

    if counts['failedNodeCount']:
        retcode = _RC_FAIL

    if (unverified and unverified_lst) or output:
        if not output:
            if not jsonify:
                print utils.color_wrap(BgColors.GREEN,
                                       'Checked Nodes: %s, '
                                       'Checked Links: %s, '
                                       'Failed Nodes: %s, '
                                       'Failed Links: %s' %
                                       (counts['checkedNodeCount'],
                                        counts['totalLinkCount'],
                                        counts['failedNodeCount'],
                                        counts['failedLinkCount']))
                print 'No VLAN or PVID Mismatch found'
        else:
            if not jsonify:
                print utils.color_wrap(BgColors.RED,
                                       'Checked Nodes: %s, '
                                       'Checked Links: %s, '
                                       'Failed Nodes: %s, '
                                       'Failed Links: %s' %
                                       (counts['checkedNodeCount'],
                                        counts['totalLinkCount'],
                                        counts['failedNodeCount'],
                                        counts['failedLinkCount']))
                if output:
                    print utils.color_wrap(BgColors.RED,
                                           'Vlan and/or PVID mismatch found on'
                                           ' following links')
            _tabulate(output,
                      headers=['Node', 'Interface', 'Vlans',
                               'Peer', 'Peer Interface', 'Peer Vlans', 'Error'],
                      jsonify=jsonify)
            # We have to convert the JSON above into something more
            #  and save it since we have multiple prints and we
            # want to incorporate them all into one big JSON string
            if jsonify:
                out_json['failedNodes'] = _JSON[0]

        if unverified and unverified_lst:
            if not jsonify:
                print utils.color_wrap(BgColors.RED,
                                       'Vlan check could not be done on:')
            _tabulate(unverified_lst,
                      headers=['Node', 'Interface', 'Vlans', 'Peer',
                               'Peer Interface', 'Reason'],
                      jsonify=jsonify)
            if jsonify:
                out_json['unverifiedLinks'] = _JSON[0]
    else:
        if not jsonify:
            print utils.color_wrap(BgColors.GREEN,
                                   'Checked Nodes: %s, '
                                   'Checked Links: %s, '
                                   'Failed Nodes: %s, '
                                   'Failed Links: %s' %
                                   (counts['checkedNodeCount'],
                                    counts['totalLinkCount'],
                                    counts['failedNodeCount'],
                                    counts['failedLinkCount']))
            print 'No VLAN or PVID Mismatch found'

    if jsonify:
        _JSON = out_json

    return retcode


def cli_check_mtu(cli, netq):
    '''Perform MTU consistency check'''

    global _JSON  # pylint: disable=global-statement

    jsonify = cli.get('json')

    hostname = cli.get('<hostname>') or '*'
    ifname = cli.get('<remote-interface>')  # we want this to be not '*'
    unverified = cli.get('unverified')
    retcode = _RC_SUCCESS

    unverified_lst, counts, output = (
        netq.check_mtu_consistency(hostname, ifname))

    out_json = {'failedNodes': [],
                'unverifiedLinks': [],
                'summary': counts}

    if counts['failedNodeCount']:
        retcode = _RC_FAIL

    if (unverified and unverified_lst) or output:
        if not output:
            if not jsonify:
                print utils.color_wrap(BgColors.GREEN,
                                       'Checked Nodes: %s, '
                                       'Checked Links: %s, '
                                       'Failed Nodes: %s, '
                                       'Failed Links: %s' %
                                       (counts['checkedNodeCount'],
                                        counts['totalLinkCount'],
                                        counts['failedNodeCount'],
                                        counts['failedLinkCount']))
                print 'No links with MTU mismatch found'
        else:
            if not jsonify:
                print utils.color_wrap(BgColors.RED,
                                       'Checked Nodes: %s, '
                                       'Checked Links: %s, '
                                       'Failed Nodes: %s, '
                                       'Failed Links: %s' %
                                       (counts['checkedNodeCount'],
                                        counts['totalLinkCount'],
                                        counts['failedNodeCount'],
                                        counts['failedLinkCount']))
                if output:
                    print utils.color_wrap(BgColors.RED,
                                           'MTU mismatch found on'
                                           ' following links')
            _tabulate(output,
                      headers=['Node', 'Interface', 'MTU',
                               'Peer', 'Peer Interface', 'Peer MTU', 'Error'],
                      jsonify=jsonify)
            # We have to convert the JSON above into something more
            #  and save it since we have multiple prints and we
            # want to incorporate them all into one big JSON string
            if jsonify:
                out_json['failedNodes'] = _JSON[0]

        if unverified and unverified_lst:
            if not jsonify:
                print utils.color_wrap(BgColors.RED,
                                       'MTU check could not be done on:')
            _tabulate(unverified_lst,
                      headers=['Node', 'Interface', 'MTU', 'Peer',
                               'Peer Interface', 'Reason'],
                      jsonify=jsonify)
            if jsonify:
                out_json['unverifiedLinks'] = _JSON[0]
    else:
        if not jsonify:
            print utils.color_wrap(BgColors.GREEN,
                                   'Checked Nodes: %s, '
                                   'Checked Links: %s, '
                                   'Failed Nodes: %s, '
                                   'Failed Links: %s' %
                                   (counts['checkedNodeCount'],
                                    counts['totalLinkCount'],
                                    counts['failedNodeCount'],
                                    counts['failedLinkCount']))
            print 'No MTU Mismatch found'

    if jsonify:
        _JSON = out_json

    return retcode


def cli_check_bgp(cli, netq):
    '''CLI BGP check handling'''

    global _JSON

    jsonify = cli.get('json')

    vrf = cli.get('<vrf>') or '*'

    good_out, counts, out = netq.check_bgp_status(vrf)
    retcode = _RC_SUCCESS

    out_json = {'failedNodes': [],
                'summary': counts}

    if counts['failedNodeCount']:
        retcode = _RC_FAIL

    col_headers = ['Node', 'Neighbor', 'Peer ID', 'Reason', 'Time']
    if counts['failedSessionCount'] or counts['failedNodeCount']:
        if not jsonify:
            print utils.color_wrap(BgColors.RED,
                                   'Total Nodes: %d, '
                                   'Failed Nodes: %d, '
                                   'Total Sessions: %d , '
                                   'Failed Sessions: %d, ' %
                                   (counts['checkedNodeCount'],
                                    counts['failedNodeCount'],
                                    counts['totalSessionCount'],
                                    counts['failedSessionCount']))
        if out:
            _tabulate([(utils.color_wrap(BgColors.RED, hostname),
                        utils.color_wrap(BgColors.RED, nbr),
                        utils.color_wrap(BgColors.RED, peerid),
                        utils.color_wrap(BgColors.RED, reason),
                        utils.color_wrap(BgColors.RED,
                                         utils.pretty_date(rst_time)))
                       for hostname, nbr, peerid, reason, rst_time in
                       sorted(out)], headers=col_headers, jsonify=jsonify)
        if jsonify:
            out_json['failedNodes'] = _JSON[0]
    else:
        if not counts['totalSessionCount']:
            if not jsonify:
                print('No BGP session info found. '
                      'Total Nodes: %d, Failed Nodes: 0' % (
                          counts['checkedNodeCount']))

        elif not jsonify:
            print utils.color_wrap(
                BgColors.GREEN,
                'Total Nodes: %d, Failed Nodes: 0, Total Sessions: %d, '
                'Failed Sessions: 0' % (counts['checkedNodeCount'],
                                        counts['totalSessionCount']))
    if jsonify:
        _JSON = out_json

    return(retcode)


def cli_check_ospf(cli, netq):
    '''CLI check OSPF status'''

    global _JSON

    jsonify = cli.get('json')
    retcode = _RC_SUCCESS

    out, counts = netq.check_ospf_status()
    out_json = {'failedNodes': [],
                'summary': counts}

    col_headers = ['Node', 'Interface', 'PeerID', 'IP Addr', 'Reason',
                   'Timestamp']

    if not counts['checkedNodeCount']:
        if not jsonify:
            print 'No OSPF session info found'
    elif counts['failedAdjCount'] or counts['failedNodeCount']:
        retcode = _RC_FAIL
        if not jsonify:
            print utils.color_wrap(BgColors.RED,
                                   'Total Nodes: %d, '
                                   'Failed Nodes: %d, '
                                   'Total Sessions: %d , '
                                   'Failed Sessions: %d, ' %
                                   (counts['checkedNodeCount'],
                                    counts['failedNodeCount'],
                                    counts['totalAdjCount'],
                                    counts['failedAdjCount']))
        if out:
            _tabulate([(utils.color_wrap(BgColors.RED, hostname),
                        utils.color_wrap(BgColors.RED, nbr),
                        utils.color_wrap(BgColors.RED, peerid),
                        utils.color_wrap(BgColors.RED, addr),
                        utils.color_wrap(BgColors.RED, state),
                        utils.color_wrap(BgColors.RED,
                                         utils.pretty_date(rst_time)))
                       for hostname, nbr, peerid, addr, state, rst_time in
                       sorted(out)],
                      headers=col_headers,
                      jsonify=jsonify)

    else:
        if not jsonify:
            print utils.color_wrap(
                BgColors.GREEN,
                'Total Sessions: %d, Failed Sessions: 0' %
                (counts['totalAdjCount']))
        else:
            _tabulate(sorted(out), headers=col_headers, jsonify=jsonify)

    if jsonify:
        out_json['failedNodes'] = _JSON[0]
        _JSON = out_json

    return retcode


def cli_check_lnv(cli, netq):
    '''Check LNV across the fabric'''

    global _JSON  # pylint: disable=global-statement

    jsonify = cli.get('json')

    col_headers = ['Node', 'Reason']
    retcode = _RC_SUCCESS

    (_, bad, counts) = netq.check_lnv_status()

    out_json = {'summary': counts,
                'failedNodes': []}

    if counts['checkedNodeCount'] == 0:
        if not jsonify:
            print 'No LNV info found'
            return
    else:
        if not jsonify:
            if (counts['warningNodeCount'] == 0 and
                counts['failedNodeCount'] == 0):
                color = BgColors.GREEN
            elif counts['failedNodeCount'] != 0:
                color = BgColors.RED
                retcode = _RC_FAIL
            else:
                color = BgColors.WARNING

            print utils.color_wrap(
                color,
                'Checked Nodes: %d, Warning Nodes: %d, Failed Nodes: %d'
                % (counts['checkedNodeCount'],
                   counts['warningNodeCount'],
                   counts['failedNodeCount'])
            )

    if counts['failedNodeCount'] or counts['warningNodeCount'] or jsonify:
        new_failed_node_list = []
        for nodename, err_str in bad:
            new_failed_node_list.append(
                (nodename, utils.color_wrap(BgColors.RED, err_str)))

        colsizes = [16, 64]
        _tabulate(sorted(new_failed_node_list),
                  headers=col_headers,
                  colsizes=colsizes,
                  jsonify=jsonify)
        if jsonify:
            out_json['failedNodes'] = _JSON[0]

    if jsonify:
        _JSON = out_json

    return retcode


def cli_check_vxlan(cli, netq):
    '''Check VXLAN across the fabric'''

    global _JSON  # pylint: disable=global-statement

    jsonify = cli.get('json')

    col_headers = ['Node', 'Reason']
    colsizes = [16, 64]
    retcode = _RC_SUCCESS

    (out, warn, bad, counts) = netq.check_vxlan_status()

    out_json = {'summary': counts,
                'failedNodes': []}

    if counts['checkedNodeCount'] == 0:
        if not jsonify:
            print('No VXLAN info found')
            return
    else:
        if not jsonify:
            if ((counts['warningNodeCount'] == 0)
                    and (counts['failedNodeCount'] == 0)):
                color = BgColors.GREEN
            elif counts['failedNodeCount'] != 0:
                color = BgColors.RED
                retcode = _RC_FAIL
            else:
                color = BgColors.WARNING

            print(utils.color_wrap(
                color,
                'Checked Nodes: %d, Warning Nodes: %d, Failed Nodes: %d'
                % (counts['checkedNodeCount'],
                   counts['warningNodeCount'],
                   counts['failedNodeCount'])
            ))

    if counts['failedNodeCount']:
        new_failed_node_list = []
        if not jsonify:
            print('Nodes with error\n')
        for node_name, err_str in bad:
            new_failed_node_list.append(
                (node_name, utils.color_wrap(BgColors.RED, err_str)))
        _tabulate(sorted(new_failed_node_list),
                  headers=col_headers,
                  colsizes=colsizes,
                  jsonify=jsonify)
        if jsonify:
            out_json['failedNodes'] = _JSON[0]
            _JSON = []

    if counts['warningNodeCount']:
        new_failed_node_list = []
        if not jsonify:
            print('Nodes with Warning\n')
        for node_name, err_str in warn:
            new_failed_node_list.append((node_name, err_str))
        _tabulate(sorted(new_failed_node_list),
                  headers=col_headers,
                  colsizes=colsizes,
                  jsonify=jsonify)
        if jsonify:
            out_json['warningNodes'] = _JSON[0]
            _JSON = []

    if jsonify:
        _JSON = out_json

    return retcode


def cli_check_clag(cli, netq):
    '''Process CLI check clag'''

    global _JSON  # pylint: disable=global-statement

    jsonify = cli.get('json')

    col_headers = ['Node', 'Reason']
    retcode = _RC_SUCCESS

    (good_node_lst, warn_node_lst, failed_node_lst,
     counts) = netq.check_clag_status()

    if counts['failedNodeCount']:
        retcode = _RC_FAIL

    out_json = {'summary': counts,
                'failedNodes': []}
    if not counts['checkedNodeCount']:
        if not jsonify:
            print 'No CLAG session info found'
    elif not counts['failedNodeCount']:
        if not counts['warningNodeCount']:
            if not jsonify:
                print utils.color_wrap(
                    BgColors.GREEN,
                    'Checked Nodes: %d, Failed Nodes: 0'
                    % counts['checkedNodeCount']
                )
        else:
            if not jsonify:
                print utils.color_wrap(
                    BgColors.WARNING,
                    'Checked Nodes: %d, Warning Nodes: %d'
                    % (counts['checkedNodeCount'], counts['warningNodeCount'])
                )
    else:
        if not jsonify:
            if not counts['warningNodeCount']:
                print utils.color_wrap(
                    BgColors.RED,
                    'Checked Nodes: %d, Failed Nodes: %d ' % (
                        counts['checkedNodeCount'],
                        counts['failedNodeCount']))
            else:
                print utils.color_wrap(
                    BgColors.RED,
                    'Checked Nodes: %d, '
                    'Warning Nodes: %d, '
                    'Failed Nodes: %d ' % (counts['checkedNodeCount'],
                                           counts['warningNodeCount'],
                                           counts['failedNodeCount']))

    if counts['failedNodeCount']:
        new_failed_node_list = []
        if not jsonify:
            print 'Nodes with error\n'
        for nodename, err_str in failed_node_lst:
            if isinstance(err_str, list):
                err_str = ', '.join(err_str)
            if err_str:
                new_failed_node_list.append(
                    (nodename, utils.color_wrap(BgColors.RED, err_str)))
            else:
                new_failed_node_list.append((nodename, ''))
        colsizes = [16, 74]
        _tabulate(sorted(new_failed_node_list),
                  headers=col_headers,
                  colsizes=colsizes,
                  jsonify=jsonify)
        if jsonify:
            out_json['failedNodes'] = _JSON[0]
            _JSON = []
    if counts['warningNodeCount']:
        if not jsonify:
            print 'Nodes with Warning\n'
        colsizes = [16, 74]
        _tabulate(sorted(warn_node_lst),
                  headers=col_headers,
                  colsizes=colsizes,
                  jsonify=jsonify)
        if jsonify:
            out_json['warningNodes'] = _JSON[0]

    if jsonify:
        _JSON = out_json

    return retcode


def cli_check_agents(cli, netq):
    '''Check Netq agent status'''

    global _JSON

    out = netq.show_status()
    total = 0
    bad_node_count = 0
    jsonify = cli.get('json')
    retcode = _RC_SUCCESS

    bad_nodes = []
    for nodename, status, uptime, ts1 in out:
        if status == AgentHealth.ROTTEN:
            new_status = utils.color_wrap(BgColors.RED, 'Rotten')
            bad_node_count += 1
            if not jsonify:
                ts2 = datetime.isoformat(datetime.utcfromtimestamp(int(ts1)), ' ')
            else:
                ts2 = ts1
            bad_nodes.append([utils.color_wrap(BgColors.RED, nodename),
                              utils.color_wrap(BgColors.RED, 'Rotten'),
                              utils.color_wrap(BgColors.RED, str(ts2))])
            retcode = _RC_FAIL

        total += 1
    if not jsonify:
        if bad_node_count:
            print utils.color_wrap(
                BgColors.RED,
                'Checked nodes: %s, Rotten nodes: %d' % (total, bad_node_count))
        else:
            print utils.color_wrap(
                BgColors.GREEN,
                'Checked nodes: %s, Rotten nodes: %d' % (total, bad_node_count))

    if bad_nodes:
        _tabulate(sorted(bad_nodes),
                  headers=['Node', 'Status', 'Connect Time'],
                  jsonify=jsonify)
        if jsonify:
            _JSON = {'failedNodes': _JSON[0],
                     'failedNodeCount': bad_node_count,
                     'checkedNodeCount': total}
    elif jsonify:
        _JSON = {'failedNodes': [],
                 'failedNodeCount': 0,
                 'checkedNodeCount': total}

    return retcode


def cli_check_sensors(cli, netq):
    '''Check sensors CLI'''

    global _JSON  # pylint: disable=global-statement
    hostname = cli.get('<hostname>') or '*'
    jsonify = cli.get('json')
    good_out, counts, out = netq.check_sensors(hostname)
    out_json = {'failedNodes': [],
                'summary': counts}
    col_headers = ['Node', 'Name', 'Description', 'State', 'Value', 'Message']

    if not counts['checkedSensorCount']:
        if not jsonify:
            if counts['failedNodeCount']:
                print utils.color_wrap(
                    BgColors.RED,
                    'No Sensor info found. Total Nodes: %d, Failed Nodes: %d' %
                    (counts['checkedNodeCount'], counts['failedNodeCount']))
            else:
                print('No Sensor info found. Total Nodes: %d, Failed Nodes: %d'
                      % (counts['checkedNodeCount'], counts['failedNodeCount']))

    elif counts['failedSensorCount'] or counts['failedNodeCount']:
        if not jsonify:
            print utils.color_wrap(
                BgColors.RED, 'Total Nodes: %d, '
                'Failed Nodes: %d, Checked Sensors: %d, Failed Sensors: %d' %
                (counts['checkedNodeCount'], counts['failedNodeCount'],
                 counts['checkedSensorCount'], counts['failedSensorCount']))
        if out:
            _tabulate([(utils.color_wrap(BgColors.RED, hostname),
                       utils.color_wrap(BgColors.RED, name),
                       utils.color_wrap(BgColors.RED, desc),
                       utils.color_wrap(BgColors.RED, state),
                       utils.color_wrap(BgColors.RED, str(value)),
                       utils.color_wrap(BgColors.RED, msg))
                       for hostname, name, desc, state, value, msg in
                       sorted(out)], headers=col_headers, jsonify=jsonify)
        if jsonify:
            out_json['failedNodes'] = _JSON[0]
    else:
        if not jsonify:
            print utils.color_wrap(
                BgColors.GREEN,
                'Total Nodes: %d, Failed Nodes: 0, Checked Sensors: %d, '
                'Failed Sensors: 0' % (counts['checkedNodeCount'],
                                       counts['checkedSensorCount']))
    if jsonify:
        _JSON = out_json

    return (_RC_FAIL
            if (counts['failedSensorCount'] or counts['failedNodeCount'])
            else _RC_SUCCESS)


def cli_check_docker_swarm_cluster(cli, netq):
    '''check docker swarm cluster'''
    global _JSON  # pylint: disable=global-statement
    hostname = cli.get('<hostname>') or '*'
    cluster_id = cli.get('<cluster-id>') or '*'
    cluster_name = cli.get('<cluster-name>') or '*'
    jsonify = cli.get('json')
    headers = ['Check', 'Status', 'Reason', 'Description']
    rc = _RC_SUCCESS

    out, counts = netq.check_docker_swarm_cluster(
        hostname, cluster_id=cluster_id, cluster_name=cluster_name)

    if not out:
        if not jsonify:
            print 'No Swarm Cluster information found'
        else:
            out_json = {}
            out_json['summary'] = {
                'checkedNodeCount': len(counts.get('checkedNodes')),
                'failedNodeCount': len(counts.get('failedNodes'))
            }
            _JSON = out_json
    else:
        out_list = []
        for entry in out:
            if entry[1] == 'Failed':
                out_list.append((entry[0],
                                 utils.color_wrap(BgColors.RED, entry[1]),
                                 utils.color_wrap(BgColors.RED, entry[2]),
                                 entry[3]))
        if out_list:
            _tabulate(sorted(out_list, key=lambda x: x[0]), headers=headers, jsonify=jsonify)

        if jsonify:
            out_json = {}
            out_json['failedNodes'] = counts.get('failedNodes')
            out_json['summary'] = {
                'checkedNodeCount': len(counts.get('checkedNodes')),
                'failedNodeCount': len(counts.get('failedNodes'))
            }
            # out_json['detail'] = _JSON[0]
            _JSON = out_json

    return rc


def cli_check_docker_swarm_service(cli, netq):
    '''check docker swarm cluster'''
    global _JSON  # pylint: disable=global-statement
    hostname = cli.get('<hostname>') or '*'
    service_name = cli.get('<swarm-service-name>') or '*'
    jsonify = cli.get('json')
    headers = ['Host', 'Service', 'Issue']
    rc = _RC_SUCCESS

    out, counts = netq.check_docker_service(hostname,
                                            service_name=service_name)

    if not out:
        if not jsonify:
            print 'No Docker Service mismatch found'
        else:
            out_json = {}
            out_json['summary'] = {
                'checkedNodeCount': len(counts.get('checkedNodes')),
                'failedNodeCount': len(counts.get('failedNodes'))
            }
            _JSON = out_json
    else:
        out_list = []
        for entry in out:
            out_list.append((entry[0], entry[1],
                             utils.color_wrap(BgColors.RED, entry[2])))
        if out_list:
            _tabulate(sorted(out_list, key=lambda x: x[0]), headers=headers, jsonify=jsonify)

        if jsonify:
            out_json = {}
            out_json['failedNodes'] = counts.get('failedNodes')
            out_json['summary'] = {
                'checkedNodeCount': len(counts.get('checkedNodes')),
                'failedNodeCount': len(counts.get('failedNodes'))
            }
            _JSON = out_json

    return rc


def cli_check_docker(cli, netq):
    '''Check docker CLI'''
    rc = _RC_FAIL
    swarm = cli.get('swarm') or None
    service = cli.get('service') or None

    if swarm:
        rc = cli_check_docker_swarm_cluster(cli, netq)
    elif service:
        rc = cli_check_docker_swarm_service(cli, netq)
    return rc


def cli_check_license(cli, netq):
    '''Check CL License CLI'''

    global _JSON  # pylint: disable=global-statement
    jsonify = cli.get('json')
    good_out, counts, out = netq.check_license()
    out_json = {'failedNodes': [], 'summary': counts}
    col_headers = ['Node', 'Name', 'State', 'License']

    if not counts['checkedNodeCount']:
        if not jsonify:
            print('No License info found.')
    elif not counts['failedNodeCount'] and not counts['failedLicenseCount']:
        if not jsonify:
            print(utils.color_wrap(
                BgColors.GREEN,
                'Total Nodes: %d, Failed Nodes: %d, '
                'Checked Licenses: %d, Failed Licenses: %d' %
                (counts['checkedNodeCount'], counts['failedNodeCount'],
                 counts['checkedLicenseCount'], counts['failedLicenseCount'])))
    else:
        if not jsonify:
            print(utils.color_wrap(
                BgColors.RED,
                'Total Nodes: %d, Failed Nodes: %d, '
                'Checked Licenses: %d, Failed Licenses: %d' %
                (counts['checkedNodeCount'], counts['failedNodeCount'],
                 counts['checkedLicenseCount'], counts['failedLicenseCount'])))
        _tabulate([(utils.color_wrap(BgColors.RED, hostname),
                    utils.color_wrap(BgColors.RED, name),
                    utils.color_wrap(BgColors.RED, state),
                    utils.color_wrap(BgColors.RED, lic))
                   for hostname, name, state, lic in sorted(out)],
                  headers=col_headers, jsonify=jsonify)
        if jsonify:
            out_json['failedNodes'] = _JSON[0]

    if jsonify:
        _JSON = out_json
    return (_RC_FAIL
            if (counts['failedNodeCount'] or counts['failedLicenseCount'])
            else _RC_SUCCESS)


def cli_trace_path(cli, netq):
    '''Process the trace commands in the CLI'''

    global _JSON  # pylint: disable=global-statement

    jsonify = cli.get('json')

    visited_nodes = set()

    if cli.get('<mac>'):
        output, error = netq.path_trace_mac(cli.get('<mac>'),
                                            cli.get('<1-4096>') or '*',
                                            cli.get('<hostname>'), None,
                                            cli.get('<vrf>') or 'default',
                                            True, visited_nodes)
    else:
        output, error = netq.path_trace_route(
            cli.get('<ip>'), cli.get('<hostname>'), cli.get('<ip-src>'),
            cli.get('<vrf>'), True, visited_nodes, None, None, None
        )

    if not output:
        if cli.get('l2'):
            addr = cli.get('<mac>')
        else:
            addr = cli.get('<ip>')

        from_node = cli.get('<hostname>')
        if not from_node:
            from_node = cli.get('<ip-src>')

        if not jsonify:
            print utils.color_wrap(
                BgColors.RED,
                'No path to %s from %s found' % (addr, from_node))
        else:
            _JSON = {'error': 'No path to %s from %s found' % (addr,
                                                               from_node)}
        return _RC_FAIL

    if jsonify:
        flattened_paths = []
    else:
        flattened_paths = None
    if error and not jsonify:
        print utils.color_wrap(BgColors.RED, error)
    utils.prtree(output[0], 0, True, [], flattened_paths)
    if jsonify:
        _JSON = {'totalPaths': len(flattened_paths),
                 'paths': flattened_paths}
    else:
        print
    return _RC_SUCCESS

def _view_complete(argv, doc):
    """ Hack to remove <wildcard> hacks from network docopt
    """
    view_idx = next((idx for idx, arg in enumerate(argv) if arg == 'view'),
                    None)
    if view_idx is not None and len(argv) >= view_idx + 2:
        choices = NetQ.view_complete(argv[view_idx + 1])
        aroundstring = ' around <text-time>'
        viewstring = ('   netq <hostname> view {key}')
        usage_idx = doc.find('Usage:')
        options_idx = doc.find('Options:')
        pre_usage = doc[:usage_idx]
        usage = doc[usage_idx:options_idx].rstrip('\n')
        post_usage = doc[options_idx:]
        new_usage = '%s\n%s\n%s\n' % (usage,
                                      '\n'.join([viewstring.format(key=choice)
                                                 for choice in choices]),
                                      '\n'.join([viewstring.format(key=choice) +
                                                 aroundstring
                                                 for choice in choices]))
        return pre_usage + new_usage + post_usage
    return doc


class NetqCLI(NetworkDocopt):
    def __init__(self, daemon):

        ''''''
        global __doc__

        NetworkDocopt.__init__(self, _add_dynamic_cmds(__doc__), daemon)

        # Add custom completers
        self.add_tag_matcher('<hostname>', completers.hostname_match, True)
        self.add_tab_complete_hook('<hostname>', completers.hostname_complete)
        self.add_tag_matcher('<remote-interface>', completers.interface_match,
                             True)
        self.add_tab_complete_hook('<remote-interface>',
                                   completers.interface_complete)
        self.add_tag_matcher('<remote-physical-interface>',
                             completers.interface_match, True)
        self.add_tab_complete_hook('<remote-physical-interface>',
                                   completers.physical_interface_complete)
        self.add_tag_matcher('<os-version>', completers.generic_match, True)
        self.add_tab_complete_hook('<os-version>',
                                   completers.os_version_complete)
        self.add_tag_matcher('<os-name>', completers.generic_match, True)
        self.add_tab_complete_hook('<os-name>', completers.os_name_complete)
        self.add_tag_matcher('<cpu-arch>', completers.generic_match, True)
        self.add_tab_complete_hook('<cpu-arch>', completers.cpu_arch_complete)
        self.add_tag_matcher('<cpu-model>', completers.generic_match, True)
        self.add_tab_complete_hook('<cpu-model>', completers.cpu_model_complete)
        self.add_tag_matcher('<board-vendor>', completers.generic_match, True)
        self.add_tag_matcher('<board-model>', completers.generic_match, True)
        self.add_tab_complete_hook('<board-vendor>',
                                   completers.board_vendor_complete)
        self.add_tab_complete_hook('<board-model>',
                                   completers.board_model_complete)
        self.add_tag_matcher('<asic-vendor>', completers.generic_match, True)
        self.add_tag_matcher('<asic-model>', completers.generic_match, True)
        self.add_tag_matcher('<asic-model-id>', completers.generic_match, True)
        self.add_tab_complete_hook('<asic-vendor>',
                                   completers.asic_vendor_complete)
        self.add_tag_matcher('<disk-name>', completers.generic_match, True)
        self.add_tab_complete_hook('<disk-name>',
                                   completers.disk_name_complete)
        self.add_tag_matcher('<disk-size>', completers.generic_match, True)
        self.add_tab_complete_hook('<disk-size>',
                                   completers.disk_size_complete)
        self.add_tag_matcher('<disk-transport>', completers.generic_match,
                             True)
        self.add_tab_complete_hook('<disk-transport>',
                                   completers.disk_transport_complete)
        self.add_tag_matcher('<disk-vendor>', completers.generic_match, True)
        self.add_tab_complete_hook('<disk-vendor>',
                                   completers.disk_vendor_complete)
        self.add_tag_matcher('<memory-type>', completers.generic_match, True)
        self.add_tab_complete_hook('<memory-type>',
                                   completers.memory_type_complete)
        self.add_tag_matcher('<memory-size>', completers.generic_match, True)
        self.add_tab_complete_hook('<memory-size>',
                                   completers.memory_size_complete)
        self.add_tag_matcher('<memory-vendor>', completers.generic_match, True)
        self.add_tab_complete_hook('<memory-vendor>',
                                   completers.memory_vendor_complete)
        self.add_tab_complete_hook('<asic-model>',
                                   completers.asic_model_complete)
        self.add_tab_complete_hook('<asic-model-id>',
                                   completers.asic_model_id_complete)
        self.add_tag_matcher('<temp-name>', completers.generic_match, True)
        self.add_tab_complete_hook('<temp-name>',
                                   completers.temp_name_complete)
        self.add_tag_matcher('<fan-name>', completers.generic_match, True)
        self.add_tab_complete_hook('<fan-name>', completers.fan_name_complete)
        self.add_tag_matcher('<psu-name>', completers.generic_match, True)
        self.add_tab_complete_hook('<psu-name>', completers.psu_name_complete)
        self.add_tag_matcher('<interface-nexthop>',
                             completers.generic_match, True)
        self.add_tab_complete_hook('<interface-nexthop>',
                                   completers.interface_complete)
        self.add_tag_matcher('<inventory-search>',
                             completers.inventory_match, True)
        self.add_tag_matcher('<ip-server>', completers.ip_server_match, True)
        self.add_tab_complete_hook('<vrf>', completers.vrf_complete)
        self.add_tag_matcher('<vrf>', completers.vrf_match, True)
        self.add_tab_complete_hook('<service-name>',
                                   completers.service_complete)
        self.add_tag_matcher('<service-name>', completers.service_match, True)
        self.add_tag_matcher('<server-name>', completers.server_name_match,
                             True)
        self.add_tab_complete_hook('<mode>', completers.mode_complete)
        self.add_tag_matcher('<mode>', completers.mode_match, True)
        self.add_tab_complete_hook('<network-name>',
                                   completers.network_name_complete)
        self.add_tag_matcher('<network-name>', completers.network_name_match,
                             True)
        self.add_tab_complete_hook('<swarm-service-name>',
                                   completers.swarm_service_complete)
        self.add_tag_matcher('<swarm-service-name>',
                             completers.swarm_service_match, True)
        self.add_tab_complete_hook('<proto>', completers.proto_complete)
        self.add_tag_matcher('<proto>', completers.proto_match, True)
        self.add_tag_matcher('<port>', completers.port_match, True)
        self.add_tag_matcher('<cluster-id>', completers.cluster_id_match, True)
        self.add_tab_complete_hook('<cluster-id>',
                                   completers.cluster_id_complete)
        self.add_tag_matcher('<cluster-name>', completers.cluster_name_match,
                             True)
        self.add_tab_complete_hook('<cluster-name>',
                                   completers.cluster_name_complete)
        self.add_tag_matcher('<role>', completers.role_match, True)
        self.add_tab_complete_hook('<role>', completers.role_complete)
        self.add_tag_matcher('<container-name>',
                             completers.container_name_match, True)
        self.add_tab_complete_hook('<container-name>',
                                   completers.container_name_complete)
        self.add_tag_matcher('<container-image>',
                             completers.container_image_match, True)
        self.add_tab_complete_hook('<container-image>',
                                   completers.container_image_complete)
        self.add_tag_matcher('<container-id>',
                             completers.container_id_match, True)
        self.add_tab_complete_hook('<container-id>',
                                   completers.container_id_complete)
        self.add_tag_matcher('<network-driver>',
                             completers.network_driver_match, True)
        self.add_tab_complete_hook('<network-driver>',
                                   completers.network_driver_complete)
        self.add_tag_matcher('<docker-version>',
                             completers.docker_version_match, True)
        self.add_tab_complete_hook('<docker-version>',
                                   completers.docker_version_complete)
        self.add_tag_matcher('<container-mac>',
                             completers.container_mac_match, True)
        self.add_tab_complete_hook('<container-mac>',
                                   completers.container_mac_complete)
        self.short_doc_string = '''
    netq command is in one the of the following forms:
       netq [<hostname>] (check|show) <object> <options>
       netq trace <options>
       netq resolve
       netq (agent|notifier) (<start|stop|restart|status>) [vrf <vrf>]

    Type:
       'netq show agents' to see the list of hosts monitored by netq
       'netq help verbose' to get the help on help
       'netq help list' to dump all the commands and keyword help
       'netq example' to see examples of usage and possible workflow
       'netq help <text-keywords>' ('netq help check bgp', for eg.) to see
       commands matching keywords'''
        self.verbose_doc_string = '''
       netq commands have the following format:
           netq [<hostname>] action object [options]

       [] denotes an optional parameter or keyword
       <> denotes a parameter to be specified such as an IP prefix, addr etc.

       Hitting the TAB key will automatically show the available options.
       Partial keywords are also accepted; e.g.: 'netq show ip ro 4.0.0.1'.

       action is one of:
           add         :  Add IP address or hostname of netq server
           agent       :  Start/Stop/Restart/Query status of netq agent
           check       :  Perform fabric-wide checks
           example     :  Show examples of usage and workflow
           help        :  Show usage info
           notifier    :  Start/Stop/Restart/Query status of netq notifier
           resolve     :  Annotate input with names and interesting info
           show        :  Show fabric-wide info about specified object
           trace       :  Control plane trace path across fabric

       object is what the information is requested about such as interfaces,
       routes, services, etc.

       Options are used to further qualify the output desired such as
       asking for JSON output or requesting information about a specific object
       within the main object.

       Time based displays are governed by two options:
           around            : Go back in time to around ...
           between           : specify start and end times for output desired

       Time itself is specified as either a relative time such as 1s, 10m, 1d
       etc. or precise time '%Y-%m-%dT%H:%M:%S.%fZ' or 'd:h:m:s'.
        '''

    def run(self, argv, uid, gid):
        """
        'parent' is the NetDaemon object used for checking user credentials.
        """
        global _JSON  # pylint: disable=global-statement
        global __doc__

        _JSON = None

        len_argv = len(argv)

        # Handle help separately
        # Special check for handling "netq help verbose" and "netq help legend"
        if len_argv == 1:
            print self.short_doc_string
            return 0
        elif argv[1] == 'help' or argv[-1] == 'help':
            if len_argv == 2:
                print self.short_doc_string
                return 0
            if argv[2] == 'verbose':
                print self.verbose_doc_string
                return 0
            if argv[2] == 'list':
                print __doc__
                return 0
            else:
                output = []
                print_section = False
                section = ''
                for line in __doc__.splitlines():
                    if not line:
                        continue
                    if line == 'Usage:':
                        section = 'Commands:'
                        print_section = True
                    elif line == 'Options:':
                        section = 'Keywords:'
                        print_section = True
                    prev_match = True
                    if argv[-1] == 'help':
                        for ele in argv[1:-1]:
                            if not re.search(r'\b%s\b' % ele, line):
                                prev_match = False
                                break
                    else:
                        for ele in argv[2:]:
                            if not re.search(r'\b%s\b' % ele, line):
                                prev_match = False
                                break
                    if prev_match:
                        if print_section:
                            print section
                            print_section = False
                        print line
            return 0

        NetworkDocopt.run(self, argv)  # Populates self.args
        args = self.args               # Convenience variable to avoid typing "self"

        start_time = end_time = None
        if args.get('around') or args.get('between'):
            start_time = args.get('<text-time>')
        if args.get('and'):
            end_time = args.get('<text-endtime>')
        jsonify = args.get('json')

        netq = NetQ(None, start_time, end_time)
        rc = _RC_SUCCESS
        cli = args

        # TESTED
        if args.get('example'):
            # 'example' MUST be handled first or else it breaks completely
            eg_choices, _ = get_examples()
            for choice in eg_choices:
                found = False
                for word in choice.split():
                    if args.get(word):
                        found = True
                    else:
                        found = False
                if found:
                    eg_file = choice.replace(' ', '/')

                    for example_dir in ('/usr/local/share/netq/example/',
                                        '/usr/share/netq/example/'):
                        if os.path.isfile(example_dir + eg_file):
                            with open(example_dir + eg_file, 'r') as f:
                                print '\n' + f.read()
                            break
            return

        elif args.get('add'):
            # We're writing the configs for the netq-agent here
            conf_dict = utils.get_yaml_config()
            server_addr = cli.get('<ip-server>')
            vrf = cli.get('<text-vrf>') or 'default'
            port = cli.get('<text-port>') or 6379
            try:
                port = int(port)
            except ValueError:
                print('Invalid Port. Should be a number')
                return _RC_FAIL
            if not server_addr:
                server_name = cli.get('<server-name>')
                if server_name:
                    server_addr = socket.gethostbyname(server_name)
            if not server_addr:
                print "Invalid server IP or name specified"
                return _RC_FAIL
            new_backend = {ConfKw.SERVER: server_addr,
                           ConfKw.PORT: port,
                           ConfKw.VRF: vrf}
            set_connection_settings(new_backend)
            try:
                ping_server()
            except BackendExceptions:
                print 'Unable to connect to backend server. Server not added.'
                close_connection()
                return _RC_FAIL

            conf_dict[ConfKw.BACKEND] = new_backend
            if utils.update_config(conf_dict):
                return _RC_FAIL
            NetworkDocopt.__init__(self, _add_dynamic_cmds(__doc__), True)
            return _RC_SUCCESS

        elif cli.get('notifier'):
            return cli_process_notifier(cli, netq)

        elif args.get('agent'):
            return cli_process_agent(cli, netq)

        # Everything below this requires connection to backend server
        if not netq:
            print "No connection to backend server"
            return
        if not netq.get_config_server_addr():
            print "No backend server specified"
            return

        if args.get('check'):

            if args.get('agents'):
                rc = cli_check_agents(cli, netq)

            if args.get('mtu'):
                rc = cli_check_mtu(cli, netq)

            elif args.get('vlan'):
                rc = cli_check_vlan(cli, netq)

            elif args.get('bgp'):
                rc = cli_check_bgp(cli, netq)

            elif cli.get('docker'):
                rc = cli_check_docker(cli, netq)

            elif args.get('ospf'):
                rc = cli_check_ospf(cli, netq)

            elif args.get('clag'):
                rc = cli_check_clag(cli, netq)

            elif args.get('license'):
                rc = cli_check_license(cli, netq)

            elif args.get('sensors'):
                rc = cli_check_sensors(cli, netq)

            elif args.get('vxlan'):
                rc = cli_check_vxlan(cli, netq)

            elif cli.get('lnv'):
                rc = cli_check_lnv(cli, netq)

        elif args.get('resolve'):
            vrf = args.get('<vrf>') or '*'
            if not sys.stdin.isatty():
                print netq.resolve_output(vrf)

        elif args.get('trace'):
            cli_trace_path(cli, netq)

        # elif args.get('view'):
        #     hostname = args.get('<hostname>')

        #     if args.get('keys'):
        #         _tabulate(sorted(netq.show_keys(hostname)),
        #                   ['Key', 'Oldest', 'Newest', 'Count'],
        #                   jsonify=jsonify)
        #     else:
        #         argv = self.argv_expanded
        #         # Taking the hostname into account
        #         startidx = argv.index('view') + 1
        #         if args.get('around'):
        #             endidx = argv.index('around')
        #             tsp, output = netq.view(hostname, argv[startidx:endidx])
        #         else:
        #             tsp, output = netq.view(hostname, argv[startidx:])

        #         print '\n'.join([
        #             utils.color_wrap(BgColors.RED, 'Output retrieved from %s' %
        #                              utils.pretty_date(tsp)), output])

        elif args.get('show'):
            if args.get('info'):
                raise NotImplementedError('show info is not implemented yet')

            elif args.get('agents'):
                cli_show_agent_status(cli, netq)
            elif args.get('bgp'):
                cli_show_bgp_status(cli, netq)
            elif args.get('clag'):
                cli_show_clag_status(cli, netq)
            elif cli.get('docker'):
                cli_show_docker(cli, netq)
            elif args.get('interfaces'):
                cli_show_links(cli, netq)
            elif args.get('lldp'):
                cli_show_lldp(cli, netq)
            elif args.get('lnv'):
                cli_show_lnv_status(cli, netq)

            elif args.get('macs'):
                cli_show_macs(cli, netq)

            elif args.get('addresses'):
                cli_show_addresses(cli, netq)
            elif args.get('services'):
                cli_show_services(cli, netq)

            elif args.get('sensors'):
                cli_show_sensors(cli, netq)
            elif args.get('inventory'):
                cli_show_inventory(cli, netq)
            elif args.get('stp'):
                cli_show_stp_topology(cli, netq)
            elif args.get('routes'):
                cli_show_routes(cli, netq)
            elif args.get('neighbors'):
                cli_show_ip_neighbors(cli, netq)

            # This has to be the last entry under 'show' to separate it from
            # the 'changes' option under the more specific objects
            elif args.get('changes'):
                cli_show_changes(cli, netq)

        if _JSON:
            print json.dumps(_JSON, indent=4)
        elif not _JSON and jsonify:
            print _EMPTY_JSON

        return rc
