# Copyright 2017 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

from __future__ import absolute_import
import copy

# To add a new PubSub event/redis key to monitor, add the key to PubSubEvents.
# E.g.: In nqv1#:#Link#:#p2#:#swp1#:#bcm_knet#:#0#:#swp1, 'Link' is the key.
PubSubEvents = ['Link', 'Services', 'BgpSession', 'ClagSession', 'OspfNbr',
                'OspfIf', 'PSU', 'Fan', 'Temp', 'License', 'LnvSession',
                'VxlanRemoteDest']


def get_events():
    event_lst = []
    for event in PubSubEvents:
        event_lst.append(type(event, (PubSubEvent,), {}))
    return event_lst


class PubSubEvent(object):

    def __init__(self, logger):
        self.channel_str = 'nqv1#:#%s#:*' % self.__class__.__name__
        self.notifiers = []
        self.notifier_handlers = []
        self.events = {}
        self._logger = logger

    def handler(self, msg):
        self._logger.debug('%s -> %s' % (msg['channel'], msg['data']))
        if not self.events.get(msg['channel']):
            self.events[msg['channel']] = (msg['data'], 1)
        else:
            _, cnt = self.events[msg['channel']]
            self.events[msg['channel']] = (msg['data'], cnt + 1)

    def add_notifier_handler(self, handler):
        self.notifier_handlers.append(handler)

    def call_notifier_handlers(self):
        if self.events:
            events_copy = copy.copy(self.events)
            self.events = {}
            for handler in self.notifier_handlers:
                handler(events_copy)
